from tree_sitter import Language, Parser
from .pcpp import pcmd
import sys
import os
import io

def macros_expand(code:str, defines:dict=None)->str:
    output = io.StringIO()
    input = io.StringIO(code)
    input.name = "<string>"

    if defines is None:
        defines = {}

    cmds = ["", "", "--line-directive"]
    for name, value in defines.items():
        arg = f"-D{name}"
        if value is not None:
            arg += f"={value}"
        cmds.append(arg)

    pcmd.CmdPreprocessor(cmds, input, output)
    return output.getvalue()

def _get_funcs_and_structs(node, func_defs:dict, struct_defs:dict):
    if node.grammar_name == "function_definition":
        func_name = str(node.children[1].children[0].text, encoding="utf-8")
        args = []
        arg_types = []
        for child in node.children[1].children[1].children:
            if child.grammar_name == "parameter_declaration":
                arg = {}
                if child.children[0].grammar_name in ["identifier", "type_identifier", "primitive_type"]:
                    arg["modifier"] = "in"
                    arg["type"] = str(child.children[0].text, encoding="utf-8")
                elif child.children[0].grammar_name in ["in", "inout", "out"]:
                    arg["modifier"] = str(child.children[0].text, encoding="utf-8")
                    arg["type"] = str(child.children[1].text, encoding="utf-8")

                args.append(arg)
                arg_types.append(arg["type"])

        func = \
        {
            "signature": f"{func_name}({', '.join(arg_types)})",
            "name": func_name,
            "args": args,
            "start": node.start_byte,
            "end": node.end_byte,
            "body": str(node.children[2].text, encoding="utf-8"),
            "node": node,
            "body_node": node.children[2],
            "used": False
        }
        if func_name not in func_defs:
            func_defs[func_name] = []

        func_defs[func_name].append(func)
        return

    if node.grammar_name == "struct_specifier":
        struct_name = str(node.children[1].text, encoding="utf-8")
        struct = {}
        struct["name"] = struct_name
        struct["start"] = node.start_byte
        struct["end"] = node.end_byte
        struct["used"] = False
        struct_defs[struct_name] = struct
        return

    for child in node.children:
        _get_funcs_and_structs(child, func_defs, struct_defs)

def _find_func_calls(node, func_calls:list):
    if node.grammar_name == "call_expression":
        func_call = {}
        func_call["name"] = str(node.children[0].text, encoding="utf-8")
        func_call["argc"] = 0
        for child in node.children[1].children:
            if child.grammar_name not in "(),":
                func_call["argc"] += 1

        func_calls.append(func_call)

    for child in node.children:
        _find_func_calls(child, func_calls)

def _remove_segments(content, segments):
    segments.sort(reverse=True, key=lambda x: x[0])
    result = content
    for start, end in segments:
        result = result[:start] + result[end:]
    return result

def _find_node_used_structs(node, structs_defs:dict):
    if node.grammar_name == "identifier":
        identifier_name = str(node.text, encoding="utf-8")
        if identifier_name in structs_defs:
            structs_defs[identifier_name]["used"] = True

    for child in node.children:
        _find_node_used_structs(child, structs_defs)

def _find_used_structs(func, structs_defs:dict):
    for arg in func["args"]:
        if arg["type"] in structs_defs:
            structs_defs[arg["type"]]["used"] = True

    body_node = func["body_node"]
    _find_node_used_structs(body_node, structs_defs)

glsl_parser = None
def treeshake(code:str, defines:dict=None)->str:
    global glsl_parser
    if glsl_parser is None:
        self_folder = os.path.dirname(os.path.abspath(__file__))
        GLSL_LANGUAGE = Language(self_folder + "/tree-sitter-glsl/glsl.dll", 'glsl')
        glsl_parser = Parser()
        glsl_parser.set_language(GLSL_LANGUAGE)

    code = macros_expand(code, defines)
    tree = glsl_parser.parse(bytes(code, sys.getdefaultencoding()))
    root_node = tree.root_node
    
    func_defs = {}
    structs_defs = {}
    _get_funcs_and_structs(root_node, func_defs, structs_defs)

    used_funcs = set()

    main_func = func_defs["main"][0]
    main_func["used"] = True
    func_stack = [main_func]
    _find_used_structs(main_func, structs_defs)
    while func_stack:
        parent_func = func_stack.pop()
        body_node = parent_func["body_node"]
        func_calls = []
        _find_func_calls(body_node, func_calls)
        for func_call in func_calls:
            func_name = func_call["name"]
            if func_name not in func_defs:
                continue

            funcs = func_defs[func_name]
            for func in funcs:
                if len(func["args"]) != func_call["argc"]:
                    continue

                signature = func["signature"]
                if signature in used_funcs:
                    continue
                
                func["used"] = True
                _find_used_structs(func, structs_defs)
                func_stack.append(func)
                used_funcs.add(signature)
    
    segments_to_remove = []
    for funcs in func_defs.values():
        for func in funcs:
            if not func["used"]:
                segments_to_remove.append((func["start"], func["end"]))

    return _remove_segments(code, segments_to_remove)



from .Manipulator import Manipulator
from glass.RenderHints import RenderHints

import glm
from OpenGL import GL
import copy
import datetime

class SceneRoamManipulator(Manipulator):

    def __init__(self):
        Manipulator.__init__(self)
        
        self._is_right_pressed = False
        self._right_press_global_pos = glm.vec2(0, 0)
        self._right_press_yaw = 0
        self._right_press_pitch = 0

        self._is_left_pressed = False
        self._left_press_global_pos = glm.vec2(0, 0)
        self._left_press_yaw = 0
        self._left_press_pitch = 0

        self._hide_cursor_global_pos = glm.vec2(0, 0)
        self._hide_cursor_yaw = 0
        self._hide_cursor_pitch = 0
        self._video_writer = None
        self._moving_speed = 1

    def on_mouse_pressed(self, button:Manipulator.MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2):
        if button == Manipulator.MouseButton.XButton1:
            self._moving_speed /= pow(2, 1/2)
        elif button == Manipulator.MouseButton.XButton2:
            self._moving_speed *= pow(2, 1/2)
        
        if not self.camera.screen.is_cursor_hiden:
            if button == Manipulator.MouseButton.RightButton:
                self._is_right_pressed = True
                self._right_press_global_pos = global_pos
                self._right_press_yaw = self.camera.yaw
                self._right_press_pitch = self.camera.pitch
            elif button == Manipulator.MouseButton.LeftButton:
                self._is_left_pressed = True
                self._left_press_global_pos = global_pos
                self._left_press_yaw = self.camera.yaw
                self._left_press_pitch = self.camera.pitch
                self._left_press_camera_pos = copy.deepcopy(self.camera.position)

        return False

    def on_mouse_released(self, button:Manipulator.MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2):
        if button == Manipulator.MouseButton.RightButton:
            self._is_right_pressed = False
        elif button == Manipulator.MouseButton.LeftButton:
            self._is_left_pressed = False
            if self._left_press_global_pos == global_pos:
                x = screen_pos.x
                y = screen_pos.y
                width = self.camera.screen.width()
                height = self.camera.screen.height()
                s = x/(width-1)
                t = 1 - y/(height-1)
                self.camera.lens.focus_tex_coord = glm.vec2(s, t)
                return True

        return False

    def on_mouse_moved(self, screen_pos:glm.vec2, global_pos:glm.vec2):
        if self.camera.screen.is_cursor_hiden:
            d = global_pos - self._hide_cursor_global_pos
            dx = d.x
            dy = d.y

            d_pitch = dy / self.camera.screen.height() * self.camera.fov_y
            d_yaw = dx / self.camera.screen.width() * self.camera.fov_x

            self.camera.yaw = self._hide_cursor_yaw - d_yaw
            self.camera.pitch = self._hide_cursor_pitch - d_pitch
            return True
        elif self._is_right_pressed:
            d = global_pos - self._right_press_global_pos
            dx = d.x
            dy = d.y

            d_pitch = dy / self.camera.screen.height() * self.camera.fov_y
            d_yaw = dx / self.camera.screen.width() * self.camera.fov_x
            
            self.camera.yaw = self._right_press_yaw + d_yaw
            self.camera.pitch = self._right_press_pitch + d_pitch
            return True
        elif self._is_left_pressed:
            d = global_pos - self._left_press_global_pos
            dx = d.x/100
            dy = d.y/100

            self.camera.position = self._left_press_camera_pos + self.camera.orientation * glm.vec3(-dx, 0, dy)
            return True
        
        return False

    # def on_wheel_scrolled(self, angle:glm.vec2, screen_pos:glm.vec2, global_pos:glm.vec2):
    #     n = angle.y/120
    #     scale = pow(2, n/6)
    #     if self.camera.projection_mode.value == 0:
    #         self.camera.fov /= scale
    #     else:
    #         self.camera.height /= scale

    #     return True

    def on_key_pressed(self, key:Manipulator.Key)->bool:
        if key in [Manipulator.Key.Key_Enter, Manipulator.Key.Key_Return]:
            self._hide_cursor_global_pos = self.camera.screen.hide_cursor()
            self._hide_cursor_yaw = self.camera.yaw
            self._hide_cursor_pitch = self.camera.pitch
        elif key == Manipulator.Key.Key_Escape:
            self.camera.screen.show_cursor()
        elif key == Manipulator.Key.Key_R:
            polygon_mode = self.camera.screen.render_hints.polygon_mode
            if polygon_mode in [GL.GL_FILL, RenderHints.inherit]:
                self.camera.screen.render_hints.polygon_mode = GL.GL_LINE
                self.camera.screen.render_hints.line_width = 1
            elif polygon_mode == GL.GL_LINE:
                self.camera.screen.render_hints.polygon_mode = GL.GL_POINT
                self.camera.screen.render_hints.point_size = 1.5
            elif polygon_mode == GL.GL_POINT:
                self.camera.screen.render_hints.polygon_mode = GL.GL_FILL
            return True
        elif key == Manipulator.Key.Key_O:
            self.camera.screen.SSAO.enabled = (not self.camera.screen.SSAO.enabled)
            return True
        elif key == Manipulator.Key.Key_M:
            self.camera.screen.DOF.enabled = (not self.camera.screen.DOF.enabled)
            return True
        elif key == Manipulator.Key.Key_P:
            now = datetime.datetime.now()
            file_name = "capture_" + now.strftime("%Y-%m-%d_%H-%M-%S") + ".png"
            self.camera.screen.capture(file_name)
        elif key == Manipulator.Key.Key_V:
            if self._video_writer is None:
                now = datetime.datetime.now()
                file_name = "capture_" + now.strftime("%Y-%m-%d_%H-%M-%S") + ".avi"
                self._video_writer = self.camera.screen.capture_video(file_name)
            else:
                self._video_writer.stop()
                self._video_writer = None
        
        return False

    def on_key_repeated(self, keys:set)->bool:
        if self._is_left_pressed:
            return

        d = self._moving_speed / 60
        fps = self.camera.screen.smooth_fps
        if fps > 0:
            d = self._moving_speed / fps
        if abs(d) < 1E-6:
            return False
        
        should_update = False
        if Manipulator.Key.Key_W in keys and \
           Manipulator.Key.Key_S not in keys: # 前进
            self.camera.position += self.camera.orientation * glm.vec3(0, d, 0)
            should_update = True
        if Manipulator.Key.Key_S in keys and \
            Manipulator.Key.Key_W not in keys: # 后退
            self.camera.position += self.camera.orientation * glm.vec3(0, -d, 0)
            should_update = True
        if Manipulator.Key.Key_A in keys and \
            Manipulator.Key.Key_D not in keys: # 向左
            self.camera.position += self.camera.orientation * glm.vec3(-d, 0, 0)
            should_update = True
        if Manipulator.Key.Key_D in keys and \
            Manipulator.Key.Key_A not in keys: # 向右
            self.camera.position += self.camera.orientation * glm.vec3(d, 0, 0)
            should_update = True
        if Manipulator.Key.Key_E in keys and \
            Manipulator.Key.Key_C not in keys: # 向上
            self.camera.position += self.camera.orientation * glm.vec3(0, 0, d)
            should_update = True
        if Manipulator.Key.Key_C in keys and \
            Manipulator.Key.Key_E not in keys: # 向下
            self.camera.position += self.camera.orientation * glm.vec3(0, 0, -d)
            should_update = True
        if Manipulator.Key.Key_F in keys:
            screen = self.camera.screen
            print("fps:", screen.fps, "draw calls:", screen.draw_calls)
            should_update = True

        return should_update


from PyQt6.QtWidgets import QWidget, QFileDialog
from PyQt6.QtGui import QColor, QMouseEvent, QPaintEvent, QPainter, QPixmap, QPen
from PyQt6.QtCore import pyqtSignal, QRect, Qt

import os

class ImageChooser(QWidget):

    file_path_changed = pyqtSignal(str)

    def __init__(self, file_path:str="", width:int=50, height:int=50, parent:QWidget=None):
        QWidget.__init__(self, parent)
        self.__file_path = file_path
        self.setFixedSize(width, height)

    @property
    def file_path(self):
        return self.__file_path
    
    @file_path.setter
    def file_path(self, file_path:str):
        if not os.path.isfile(file_path):
            return
        
        if os.path.abspath(self.__file_path).replace("\\", "/") != os.path.abspath(file_path).replace("\\", "/"):
            self.__file_path = file_path
            self.update()
            self.file_path_changed.emit(self.__file_path)

    def mouseReleaseEvent(self, mouse_event: QMouseEvent | None) -> None:
        directory = None
        if os.path.isfile(self.__file_path):
            directory = os.path.dirname(self.__file_path)

        file_path = QFileDialog.getOpenFileName(self, "选择图片", directory, "Image File(*.jpg *.png *.bmp *.ico *.exr *.hdr *.glsl)")
        if file_path[0]:
            self.file_path = file_path[0]

        return super().mouseReleaseEvent(mouse_event)
    
    def paintEvent(self, paint_event: QPaintEvent | None) -> None:
        painter = QPainter(self)
        painter.setPen(QPen(Qt.PenStyle.NoPen))
        painter.setBrush(QColor(0,0,0))
        rect = self.rect()
        painter.drawRect(rect)
        if self.isEnabled() and os.path.isfile(self.__file_path):
            pixmap = QPixmap(self.__file_path)
            pixmap_ratio = pixmap.width() / pixmap.height()
            rect_ratio = rect.width() / rect.height()
            if pixmap_ratio > rect_ratio:
                pixmap_width = rect.width()
                pixmap_height = int(rect.width() / pixmap_ratio)
                contain_rect = QRect(0, int(rect.height()/2-pixmap_height/2), pixmap_width, pixmap_height)
                painter.drawPixmap(contain_rect, pixmap)
            else:
                pixmap_height = rect.height()
                pixmap_width = int(rect.height() * pixmap_ratio)
                contain_rect = QRect(int(rect.width()/2-pixmap_width/2), 0, pixmap_width, pixmap_height)
                painter.drawPixmap(contain_rect, pixmap)

        return super().paintEvent(paint_event)


from glass.download import pip_install
from .QtScreen import init_QtScreen
from ..Manipulators.Manipulator import Manipulator

try:
    import PySide2
    from PySide2.QtOpenGLWidgets import QOpenGLWidget
except ModuleNotFoundError:
    pip_install("PySide2")
    import PySide2
    from PySide2.QtOpenGLWidgets import QOpenGLWidget

import glm

@init_QtScreen
class PySide2Screen(QOpenGLWidget):
    mouse_pressed = PySide2.QtCore.Signal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_released = PySide2.QtCore.Signal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_double_clicked = PySide2.QtCore.Signal(Manipulator.MouseButton, glm.vec2, glm.vec2)
    mouse_moved = PySide2.QtCore.Signal(glm.vec2, glm.vec2)
    wheel_scrolled = PySide2.QtCore.Signal(glm.vec2, glm.vec2, glm.vec2)
    key_pressed = PySide2.QtCore.Signal(Manipulator.Key)
    key_released = PySide2.QtCore.Signal(Manipulator.Key)
    key_repeated = PySide2.QtCore.Signal(set)
    frame_started = PySide2.QtCore.Signal()
    frame_ended = PySide2.QtCore.Signal()



from .ShaderEffect import ShaderEffect
import os

class MulFilter(ShaderEffect):

    def __init__(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        ShaderEffect.__init__(self, self_folder + "/../glsl/PostProcessEffects/mul_filter.glsl")



import numpy as np
from OpenGL import GL

from .FBOAttachment import FBOAttachment
from .GLInfo import GLInfo
from .GLConfig import GLConfig
from .GlassConfig import GlassConfig
from .BO import BO
from .samplerCube import samplerCube
from .sampler2DArray import sampler2DArray
from .sampler2D import sampler2D
from .isampler2D import isampler2D
from .usampler2D import usampler2D
from .sampler2DMS import sampler2DMS
from .isampler2DMS import isampler2DMS
from .usampler2DMS import usampler2DMS
from .RBO import RBO
from .utils import checktype
from .helper import get_external_format, get_channels

class FBO(BO):
	COLOR = GL.GL_COLOR_ATTACHMENT0
	DEPTH = GL.GL_DEPTH_ATTACHMENT
	STENCIL = GL.GL_STENCIL_ATTACHMENT
	DEPTH_STENCIL = GL.GL_DEPTH_STENCIL_ATTACHMENT

	_basic_info = \
	{
		"gen_func": GL.glGenFramebuffers,
		"bind_func": GL.glBindFramebuffer,
		"del_func": GL.glDeleteFramebuffers,
		"target_type": GL.GL_FRAMEBUFFER,
		"binding_type": GL.GL_FRAMEBUFFER_BINDING,
		"need_number": True,
	}

	_attachment_types = (sampler2D, sampler2DMS, isampler2D, isampler2DMS, usampler2D, usampler2DMS, samplerCube, sampler2DArray, RBO)
	_resolve_type_map = \
	{
		sampler2DMS: sampler2D,
		isampler2DMS: isampler2D,
		usampler2DMS: usampler2D
	}

	@checktype
	def __init__(self, width:int=0, height:int=0, samples:int=None, layers:int=None):
		BO.__init__(self, context_shared=False)
		self._last_active_id = 0
		self._resolved_fbo = None
		self._content_changed = False

		self._width = width
		self._height = height
		self._samples = samples
		self._layers = layers

		self._color_attachments = {}
		self._depth_attachment = None
		self._stencil_attachment = None
		self._depth_stencil_attachment = None

		self._attachments_attached = False
		self._can_resize = True
		self._auto_clear = True

	@checktype
	def attach(self, attach_point:int, attachment, internal_format:GLInfo.internal_formats=None):
		
		if self._attachments_attached:
			raise RuntimeError("attempt to attach new attachment to used FBO")
		
		if not (0 <= attach_point < GLConfig.max_color_attachments) and \
		   not (GL.GL_COLOR_ATTACHMENT0 <= attach_point < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments) and \
		   attach_point not in GLInfo.none_color_attachment_types:
			max_value = GLConfig.max_color_attachments-1
			max_enum = GLInfo.enum_map[GL.GL_COLOR_ATTACHMENT0 + max_value]
			error_message = f"attach target(given as {attach_point}) should be in 0~{max_value} or GL_COLOR_ATTACHMENT0~{max_enum} for color attachment\n"
			error_message += f"or be one of {GLInfo.none_color_attachment_types} for other attachments."
			raise ValueError(error_message)
		
		if GL.GL_COLOR_ATTACHMENT0 <= attach_point < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
			attach_point = attach_point - GL.GL_COLOR_ATTACHMENT0

		attachment_type = None
		if isinstance(attachment, type):
			attachment_type = attachment
			if attachment_type not in FBO._attachment_types:
				raise TypeError(f"only {FBO._attachment_types} can be attached. {attachment} was given.")
			attachment = None
		elif not isinstance(attachment, FBOAttachment):
			raise TypeError(f"only instance of FBOAttachment can be attached. {type(attachment)} value was given.")
		
		if GlassConfig.debug:
			if self._samples is not None and \
			attachment_type not in (sampler2DMS, isampler2DMS, usampler2DMS, RBO) and \
			not isinstance(attachment, (sampler2DMS, isampler2DMS, usampler2DMS, RBO)):
				raise TypeError("can only attach (sampler2DMS, isampler2DMS, usampler2DMS, RBO) to multi-samples FBO")
			
			if self._layers is not None and \
			attachment_type not in (sampler2DArray,) and \
			not isinstance(attachment, (sampler2DArray,)):
				raise TypeError(f"can only attach (sampler2DArray,) to multi-layers FBO, {attachment_type.__class__.__name__} were given.")

		if attachment is None:
			if internal_format is None:
				if attach_point == GL.GL_DEPTH_ATTACHMENT:
					internal_format = GL.GL_DEPTH_COMPONENT
				elif attach_point == GL.GL_STENCIL_ATTACHMENT:
					internal_format = GL.GL_STENCIL_INDEX
				elif attach_point == GL.GL_DEPTH_STENCIL_ATTACHMENT:
					internal_format = GL.GL_DEPTH24_STENCIL8
				elif attachment_type in [sampler2D, sampler2DMS, samplerCube, sampler2DArray, RBO]:
					internal_format = GL.GL_RGBA32F
				elif attachment_type in [isampler2D, isampler2DMS]:
					internal_format = GL.GL_RGBA32I
				elif attachment_type in [usampler2D, usampler2DMS]:
					internal_format = GL.GL_RGBA32UI

			attachment = attachment_type(internal_format=internal_format)
			if isinstance(attachment, (sampler2D, isampler2D, usampler2D, samplerCube, sampler2DArray)):
				attachment.wrap = GL.GL_CLAMP_TO_EDGE
				attachment.filter_mipmap = None
			internal_format = attachment.internal_format
		else:
			if internal_format is None:
				internal_format = attachment.internal_format

		if GlassConfig.debug:
			if attach_point == GL.GL_DEPTH_ATTACHMENT:
				if internal_format not in GLInfo.depth_internal_formats:
					raise ValueError(f"depth attachment internal format should be in {GLInfo.depth_internal_formats}, {internal_format} was given.")
			elif attach_point == GL.GL_STENCIL_ATTACHMENT:
				if internal_format not in GLInfo.stencil_internal_formats:
					raise ValueError(f"stencil attachment internal format should be in {GLInfo.depth_internal_formats}, {internal_format} was given.")
			elif attach_point == GL.GL_DEPTH_STENCIL_ATTACHMENT:
				if internal_format not in GLInfo.depth_stencil_internal_formats:
					raise ValueError(f"depth_stencil attachment internal format should be in {GLInfo.depth_stencil_internal_formats}, {internal_format} was given.")
			elif internal_format not in GLInfo.color_internal_formats:
				raise ValueError(f"color attachment internal format should be in {GLInfo.color_internal_formats}, {internal_format} was given.")

		attachment._fbo = self
		attachment._fbo_attach_point = attach_point
		attachment.malloc(width=self._width, height=self._height, samples=self._samples, layers=self._layers, internal_format=internal_format)
		if attach_point == GL.GL_DEPTH_ATTACHMENT:
			self._depth_attachment = attachment
		elif attach_point == GL.GL_STENCIL_ATTACHMENT:
			self._stencil_attachment = attachment
		elif attach_point == GL.GL_DEPTH_STENCIL_ATTACHMENT:
			self._depth_stencil_attachment = attachment
		else:
			self._color_attachments[attach_point] = attachment

	@property
	def width(self):
		return self._width
	
	@property
	def height(self):
		return self._height
	
	@property
	def samples(self):
		return self._samples

	def resize(self, width:int, height:int, samples:int=None, layers:int=None):
		if not self._can_resize:
			return
		
		if samples is None:
			samples = self._samples

		if layers is None:
			layers = self._layers
		
		self._can_resize = False

		if self._width == width and \
		   self._height == height and \
		   self._samples == samples and \
		   self._layers == layers:
			self._can_resize = True
			return False

		self._width = width
		self._height = height
		self._samples = samples
		self._layers = layers
		
		for attachment in self._color_attachments.values():
			attachment.resize(width, height, samples, layers)

		if self._depth_attachment is not None:
			self._depth_attachment.resize(width, height, samples, layers)

		if self._stencil_attachment is not None:
			self._stencil_attachment.resize(width, height, samples, layers)

		if self._depth_stencil_attachment is not None:
			self._depth_stencil_attachment.resize(width, height, samples, layers)

		self._can_resize = True
		return True

	def __enter__(self):
		self.bind()

	def __exit__(self, *exc_details):
		self.unbind()

	def bind(self):
		if self._width == 0 or self._height == 0:
			raise ValueError("current FBO is empty")

		self._last_active_id = FBO.active_id
		self._old_viewport = GLConfig.viewport

		if not self._attachments_attached:
			has_attachments = False
			color_locations = []
			max_color_location = 0
			for key, attachment in self._color_attachments.items():
				color_location = key
				if color_location > max_color_location:
					max_color_location = color_location

				BO.bind(self)
				attachment.bind(update_fbo=True)
				if isinstance(attachment, (sampler2D,isampler2D,usampler2D)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0+color_location, GL.GL_TEXTURE_2D, attachment.id, 0)
				elif isinstance(attachment, (sampler2DMS,isampler2DMS,usampler2DMS)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0+color_location, GL.GL_TEXTURE_2D_MULTISAMPLE, attachment.id, 0)
				elif isinstance(attachment, (samplerCube,sampler2DArray)):
					GL.glFramebufferTexture(GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0+color_location, attachment.id, 0)
				elif isinstance(attachment, RBO):
					GL.glFramebufferRenderbuffer(GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0+color_location, GL.GL_RENDERBUFFER, attachment.id)
				else:
					raise TypeError(f"not support attachment type {type(attachment).__name__}")
				color_locations.append(color_location)

				has_attachments = True

			if self._depth_attachment is not None:
				BO.bind(self)
				self._depth_attachment.bind(update_fbo=True)
				if isinstance(self._depth_attachment, (sampler2D,isampler2D,usampler2D)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_ATTACHMENT, GL.GL_TEXTURE_2D, self._depth_attachment.id, 0)
				elif isinstance(self._depth_attachment, (sampler2DMS,isampler2DMS,usampler2DMS)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_ATTACHMENT, GL.GL_TEXTURE_2D_MULTISAMPLE, self._depth_attachment.id, 0)
				elif isinstance(self._depth_attachment, (samplerCube,sampler2DArray)):
					GL.glFramebufferTexture(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_ATTACHMENT, self._depth_attachment.id, 0)
				elif isinstance(self._depth_attachment, RBO):
					GL.glFramebufferRenderbuffer(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_ATTACHMENT, GL.GL_RENDERBUFFER, self._depth_attachment.id)
				else:
					raise TypeError(type(self._depth_attachment))

				has_attachments = True

			if self._stencil_attachment is not None:
				BO.bind(self)
				self._stencil_attachment.bind(update_fbo=True)
				if isinstance(self._stencil_attachment, (sampler2D,isampler2D,usampler2D)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_STENCIL_ATTACHMENT, GL.GL_TEXTURE_2D, self._stencil_attachment.id, 0)
				elif isinstance(self._stencil_attachment, (sampler2DMS,isampler2DMS,usampler2DMS)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_STENCIL_ATTACHMENT, GL.GL_TEXTURE_2D_MULTISAMPLE, self._stencil_attachment.id, 0)
				elif isinstance(self._stencil_attachment, (samplerCube,sampler2DArray)):
					GL.glFramebufferTexture(GL.GL_FRAMEBUFFER, GL.GL_STENCIL_ATTACHMENT, self._stencil_attachment.id, 0)
				elif isinstance(self._stencil_attachment, RBO):
					GL.glFramebufferRenderbuffer(GL.GL_FRAMEBUFFER, GL.GL_STENCIL_ATTACHMENT, GL.GL_RENDERBUFFER, self._stencil_attachment.id)
				else:
					raise TypeError(type(self._stencil_attachment))

				has_attachments = True

			if self._depth_stencil_attachment is not None:
				BO.bind(self)
				self._depth_stencil_attachment.bind(update_fbo=True)
				if isinstance(self._depth_stencil_attachment, (sampler2D,isampler2D,usampler2D)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_STENCIL_ATTACHMENT, GL.GL_TEXTURE_2D, self._depth_stencil_attachment.id, 0)
				elif isinstance(self._depth_stencil_attachment, (sampler2DMS,isampler2DMS,usampler2DMS)):
					GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_STENCIL_ATTACHMENT, GL.GL_TEXTURE_2D_MULTISAMPLE, self._depth_stencil_attachment.id, 0)
				elif isinstance(self._depth_stencil_attachment, (samplerCube,sampler2DArray)):
					GL.glFramebufferTexture(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_STENCIL_ATTACHMENT, self._depth_stencil_attachment.id, 0)
				elif isinstance(self._depth_stencil_attachment, RBO):
					GL.glFramebufferRenderbuffer(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_STENCIL_ATTACHMENT, GL.GL_RENDERBUFFER, self._depth_stencil_attachment.id)
				else:
					raise TypeError(type(self._depth_stencil_attachment))

				has_attachments = True

			if not has_attachments:
				raise RuntimeError("FBO " + str(self._id) + " do not have any attachments")

			if color_locations:
				color_attach_points = []
				for i in range(max_color_location+1):
					if i in color_locations:
						color_attach_points.append(GL.GL_COLOR_ATTACHMENT0 + i)
					else:
						color_attach_points.append(GL.GL_NONE)
				GL.glDrawBuffers(len(color_attach_points), np.array(color_attach_points))
			else:
				GL.glDrawBuffer(GL.GL_NONE)
				GL.glReadBuffer(GL.GL_NONE)

			if GlassConfig.debug:
				self.check_status()

			self._attachments_attached = True
		else:
			for attachment in self._color_attachments.values():
				attachment.bind(update_fbo=True)
			if self._depth_attachment is not None:
				self._depth_attachment.bind(update_fbo=True)
			if self._stencil_attachment is not None:
				self._stencil_attachment.bind(update_fbo=True)
			if self._depth_stencil_attachment is not None:
				self._depth_stencil_attachment.bind(update_fbo=True)

		BO.bind(self)
		GLConfig.viewport = (0, 0, self._width, self._height)
		if self._auto_clear:
			GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT)
		
		self._content_changed = True

	def check_status(self):
		status = GL.glCheckFramebufferStatus(GL.GL_FRAMEBUFFER)
		if status == GL.GL_FRAMEBUFFER_COMPLETE:
			return
		
		error_message = status.__repr__() + ": "
		messages = GLInfo.fbo_status_errors[status]
		if isinstance(messages, list):
			error_message += "Maybe one of following reasons:\n" + "\n".join(messages)
		else:
			error_message += messages

		raise RuntimeError(error_message)

	def _unbind(self):		
		GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self._last_active_id)
		self._last_active_id = self._id

	@property
	def auto_clear(self):
		return self._auto_clear
	
	@auto_clear.setter
	@checktype
	def auto_clear(self, flag:bool):
		self._auto_clear = flag

	def unbind(self):
		self._unbind()
		GLConfig.viewport = self._old_viewport
	
	def copy_from_active(self, *targets):
		if self.id == FBO.active_id:
			return

		targets = list(targets)
		if not targets:
			targets = [*self._color_attachments.keys(), GL.GL_DEPTH_ATTACHMENT, GL.GL_STENCIL_ATTACHMENT, GL.GL_DEPTH_STENCIL_ATTACHMENT]
		
		for i, target in enumerate(targets):
			if GL.GL_COLOR_ATTACHMENT0 <= target < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
				targets[i] = target - GL.GL_COLOR_ATTACHMENT0

		viewport = GLConfig.viewport
		with self:
			GL.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, FBO.active_id)
			GL.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, self._id)

			for target in self._color_attachments:
				if target not in targets:
					continue

				GL.glReadBuffer(GL.GL_COLOR_ATTACHMENT0 + target)
				GL.glDrawBuffer(GL.GL_COLOR_ATTACHMENT0 + target)
				GL.glBlitFramebuffer(
					viewport[0], viewport[1], viewport[2], viewport[3],
					0, 0, self._width, self._height,
					GL.GL_COLOR_BUFFER_BIT, GL.GL_LINEAR
				)
				
			if self._depth_attachment is not None and GL.GL_DEPTH_ATTACHMENT in targets:
				GL.glBlitFramebuffer(
					viewport[0], viewport[1], viewport[2], viewport[3],
					0, 0, self._width, self._height,
					GL.GL_DEPTH_BUFFER_BIT, GL.GL_NEAREST
				)
			if self._stencil_attachment is not None and GL.GL_STENCIL_ATTACHMENT in targets:
				GL.glBlitFramebuffer(
					viewport[0], viewport[1], viewport[2], viewport[3],
					0, 0, self._width, self._height,
					GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
				)
			if self._depth_stencil_attachment is not None and GL.GL_DEPTH_STENCIL_ATTACHMENT in targets:
				GL.glBlitFramebuffer(
					viewport[0], viewport[1], viewport[2], viewport[3],
					0, 0, self._width, self._height,
					GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
				)

	def draw_to_active(self, *targets):
		if self.id == FBO.active_id:
			return

		targets = list(targets)
		if not targets:
			targets = [*self._color_attachments.keys(), GL.GL_DEPTH_ATTACHMENT, GL.GL_STENCIL_ATTACHMENT, GL.GL_DEPTH_STENCIL_ATTACHMENT]
		
		for i, target in enumerate(targets):
			if GL.GL_COLOR_ATTACHMENT0 <= target < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
				targets[i] = target - GL.GL_COLOR_ATTACHMENT0

		GL.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, self._id)
		GL.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, FBO.active_id)

		viewport = GLConfig.viewport
		for target in self._color_attachments:
			if target not in targets:
				continue

			GL.glReadBuffer(GL.GL_COLOR_ATTACHMENT0 + target)
			GL.glDrawBuffer(GL.GL_COLOR_ATTACHMENT0 + target)
			GL.glBlitFramebuffer(
				0, 0, self._width, self._height,
				viewport[0], viewport[1], viewport[2], viewport[3],
				GL.GL_COLOR_BUFFER_BIT, GL.GL_LINEAR
			)
			
		if self._depth_attachment is not None and GL.GL_DEPTH_ATTACHMENT in targets:
			GL.glBlitFramebuffer(
				0, 0, self._width, self._height,
				viewport[0], viewport[1], viewport[2], viewport[3],
				GL.GL_DEPTH_BUFFER_BIT, GL.GL_NEAREST
			)
		if self._stencil_attachment is not None and GL.GL_STENCIL_ATTACHMENT in targets:
			GL.glBlitFramebuffer(
				0, 0, self._width, self._height,
				viewport[0], viewport[1], viewport[2], viewport[3],
				GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
			)
		if self._depth_stencil_attachment is not None and GL.GL_DEPTH_STENCIL_ATTACHMENT in targets:
			GL.glBlitFramebuffer(
				0, 0, self._width, self._height,
				viewport[0], viewport[1], viewport[2], viewport[3],
				GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
			)

	def draw_to(self, fbo, src_targets, dest_targets=None):
		if fbo is self or fbo is None or fbo.id == self.id:
			return

		if not isinstance(src_targets, list):
			src_targets = [src_targets]

		if not src_targets:
			src_targets = [*self._color_attachments.keys(), GL.GL_DEPTH_ATTACHMENT, GL.GL_STENCIL_ATTACHMENT, GL.GL_DEPTH_STENCIL_ATTACHMENT]
		
		if dest_targets is None:
			dest_targets = src_targets

		if not isinstance(dest_targets, list):
			dest_targets = [dest_targets]

		for i, target in enumerate(src_targets):
			if GL.GL_COLOR_ATTACHMENT0 <= target < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
				src_targets[i] = target - GL.GL_COLOR_ATTACHMENT0

		for i, target in enumerate(dest_targets):
			if GL.GL_COLOR_ATTACHMENT0 <= target < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
				dest_targets[i] = target - GL.GL_COLOR_ATTACHMENT0

		with fbo:
			GL.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, self._id)
			GL.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, fbo._id)

			for target in self._color_attachments:
				if target not in src_targets:
					continue

				index = src_targets.index(target)
				dest_target = dest_targets[index]

				GL.glReadBuffer(GL.GL_COLOR_ATTACHMENT0 + target)
				GL.glDrawBuffer(GL.GL_COLOR_ATTACHMENT0 + dest_target)
				GL.glBlitFramebuffer(
					0, 0, self._width, self._height,
					0, 0, fbo.width, fbo.height,
					GL.GL_COLOR_BUFFER_BIT, GL.GL_LINEAR
				)

			if self._depth_attachment is not None and GL.GL_DEPTH_ATTACHMENT in src_targets:
				GL.glBlitFramebuffer(
					0, 0, self._width, self._height,
					0, 0, fbo.width, fbo.height,
					GL.GL_DEPTH_BUFFER_BIT, GL.GL_NEAREST
				)
			if self._stencil_attachment is not None and GL.GL_STENCIL_ATTACHMENT in src_targets:
				GL.glBlitFramebuffer(
					0, 0, self._width, self._height,
					0, 0, fbo.width, fbo.height,
					GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
				)
			if self._depth_stencil_attachment is not None and GL.GL_DEPTH_STENCIL_ATTACHMENT in src_targets:
				GL.glBlitFramebuffer(
					0, 0, self._width, self._height,
					0, 0, fbo.width, fbo.height,
					GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
				)

	@staticmethod
	def _resolve_type(attachment):
		type_attachment = type(attachment)
		if type_attachment in FBO._resolve_type_map:
			type_attachment = FBO._resolve_type_map[type_attachment]
		return type_attachment

	@property
	def resolved(self):
		if self._samples is None:
			return self

		if self._resolved_fbo is None:
			self._resolved_fbo = FBO(self._width, self._height)
			for key, attachment in self._color_attachments.items():
				type_attachment = FBO._resolve_type(attachment)
				self._resolved_fbo.attach(key, type_attachment, attachment.internal_format)
			
			if self._depth_attachment is not None:
				type_attachment = FBO._resolve_type(self._depth_attachment)
				self._resolved_fbo.attach(GL.GL_DEPTH_ATTACHMENT, type_attachment, self._depth_attachment.internal_format)

			if self._stencil_attachment is not None:
				type_attachment = FBO._resolve_type(self._stencil_attachment)
				self._resolved_fbo.attach(GL.GL_STENCIL_ATTACHMENT, type_attachment, self._stencil_attachment.internal_format)

			if self._depth_stencil_attachment is not None:
				type_attachment = FBO._resolve_type(self._depth_stencil_attachment)
				self._resolved_fbo.attach(GL.GL_DEPTH_STENCIL_ATTACHMENT, type_attachment, self._depth_stencil_attachment.internal_format)
		else:
			self._resolved_fbo.resize(self._width, self._height)
		
		if self._content_changed:
			with self._resolved_fbo:
				GL.glBindFramebuffer(GL.GL_READ_FRAMEBUFFER, self._id)
				GL.glBindFramebuffer(GL.GL_DRAW_FRAMEBUFFER, self._resolved_fbo._id)
				for key, attachment in self._color_attachments.items():
					GL.glReadBuffer(GL.GL_COLOR_ATTACHMENT0 + key)
					GL.glDrawBuffer(GL.GL_COLOR_ATTACHMENT0 + key)
					GL.glBlitFramebuffer(
						0, 0, self._width, self._height,
						0, 0, self._width, self._height,
						GL.GL_COLOR_BUFFER_BIT, GL.GL_NEAREST
					)
				if self._depth_attachment is not None:
					GL.glBlitFramebuffer(
						0, 0, self._width, self._height,
						0, 0, self._width, self._height,
						GL.GL_DEPTH_BUFFER_BIT, GL.GL_NEAREST
					)
				if self._stencil_attachment is not None:
					GL.glBlitFramebuffer(
						0, 0, self._width, self._height,
						0, 0, self._width, self._height,
						GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
					)
				if self._depth_stencil_attachment is not None:
					GL.glBlitFramebuffer(
						0, 0, self._width, self._height,
						0, 0, self._width, self._height,
						GL.GL_DEPTH_BUFFER_BIT | GL.GL_STENCIL_BUFFER_BIT, GL.GL_NEAREST
					)

			self._content_changed = False
		
		return self._resolved_fbo

	def color_attachment(self, key:int=0):
		try:
			return self._color_attachments[key]
		except:
			return self._color_attachments[key - GL.GL_COLOR_ATTACHMENT0]

	@property
	def depth_attachment(self):
		return self._depth_attachment

	@property
	def stencil_attachment(self):
		return self._stencil_attachment

	@property
	def depth_stencil_attachment(self):
		return self._depth_stencil_attachment

	def data(self, attach_point:int=0):
		if attach_point in [GL.GL_DEPTH_ATTACHMENT, GL.GL_STENCIL_ATTACHMENT, GL.GL_DEPTH_STENCIL_ATTACHMENT]:
			raise RuntimeError("cannot read depth or stencil data")

		attachment = None
		if GL.GL_COLOR_ATTACHMENT0 <= attach_point < GL.GL_COLOR_ATTACHMENT0 + GLConfig.max_color_attachments:
			attachment = self._color_attachments[attach_point - GL.GL_COLOR_ATTACHMENT0]
		else:
			attachment = self._color_attachments[attach_point]
			attach_point = GL.GL_COLOR_ATTACHMENT0 + attach_point			

		self._last_active_id = FBO.active_id
		BO.bind(self)
		GL.glBindBuffer(GL.GL_PIXEL_PACK_BUFFER, 0)
		GL.glReadBuffer(attach_point)
		GL.glPixelStorei(GL.GL_PACK_ALIGNMENT, 1)

		external_format = get_external_format(attachment.internal_format)
		dtype = attachment.dtype
		np_dtype = GLInfo.dtype_map[attachment.dtype]
		channels = get_channels(attachment.internal_format)

		result_data = None
		if channels == 1:
			result_data = np.zeros((self._height, self._width), dtype=np_dtype)
		else:
			result_data = np.zeros((self._height, self._width, channels), dtype=np_dtype)
		
		GL.glReadPixels(0, 0, self._width, self._height, external_format, dtype, result_data)
		self._unbind()

		return result_data

	@property
	def empty(self):
		return (self._width == 0 or self._height == 0)


from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
from OpenGL import GL

class Point(Mesh):

    def __init__(self, position:glm.vec3=glm.vec3(0), color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), point_size:int=5,
                 name:str=""):
        Mesh.__init__(self, primitive_type=GL.GL_POINTS, color=color, name=name)
        self.render_hints.point_size = point_size
        self.__position = position
        self.start_building()

    def build(self):
        position = self.__position
        vertices = self.vertices

        vertices[0] = Vertex(position=position, tex_coord=glm.vec3(0.5, 0.5, 0))
        del vertices[1:]

    @property
    def position(self):
        return self.__position
    
    @position.setter
    @Mesh.param_setter
    def position(self, position:glm.vec3):
        self.__position = position

    @property
    def point_size(self):
        return self.render_hints.point_size
    
    @point_size.setter
    @checktype
    def point_size(self, point_size:int):
        self.render_hints.point_size = point_size


from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass.utils import checktype, cat, md5s, relative_path, modify_time
from glass import FBO, ShaderProgram, sampler2D, GLConfig, GlassConfig, GLInfo

from OpenGL import GL
import os
import time
import glm
from datetime import datetime

class ShaderEffect(PostProcessEffect):

    __template_content = ""
    __template_filename = os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/shader_effect_template.glsl"

    @checktype
    def __init__(self, shader_path:str=None, internal_format:GLInfo.internal_formats=GL.GL_RGBA32F, generate_mipmap:bool=False):
        PostProcessEffect.__init__(self)
        
        self._dynamic = False
        self._internal_format = internal_format
        self._generate_mipmap = generate_mipmap

        self._fbo = None
        self._program = None
        self._fragment_filename = ""
        self._uniforms = {}

        self._shader_path = ""
        self._start_time = 0
        self._last_frame_time = 0
        self._frame_index = 0
        
        if shader_path is not None:
            self.shader_path = shader_path

    @property
    def fbo(self):
        if self._fbo is None:
            self._fbo = FBO()
            self._fbo.attach(0, sampler2D, self._internal_format)
        return self._fbo

    def __hash__(self):
        return id(self)

    @property
    def shader_path(self):
        return self._shader_path

    @shader_path.setter
    @checktype
    def shader_path(self, shader_path:str):
        current_file_path = os.path.dirname(os.path.abspath(__file__))
        if not os.path.isfile(shader_path) and not os.path.isabs(shader_path):
            shader_path = current_file_path + "/" + shader_path

        if not os.path.isfile(shader_path):
            raise FileNotFoundError(shader_path)

        shader_path = os.path.abspath(shader_path).replace("\\", "/")
        if shader_path == self._shader_path:
            return
        
        file_base_name = os.path.basename(shader_path)
        dest_file_name = GlassConfig.cache_folder + "/" + file_base_name + "_" + md5s(shader_path) + ".glsl"
        if modify_time(ShaderEffect.__template_filename) > modify_time(dest_file_name):
            out_file = open(dest_file_name, "w")
            out_file.write(ShaderEffect.__template(shader_path))
            out_file.close()

        self._fragment_filename = dest_file_name
        self._shader_path = shader_path

    @property
    def program(self):
        if self._fragment_filename:
            self._program = ShaderProgram()
            self._program.compile(Frame.draw_frame_vs)
            self._program.compile(self._fragment_filename, GL.GL_FRAGMENT_SHADER)
            self._fragment_filename = ""

        if self._uniforms and self._program is not None:
            for name, value in self._uniforms.items():
                self._program[name] = value
            self._uniforms.clear()

        return self._program

    def draw(self, screen_image:sampler2D)->sampler2D:
        if self._generate_mipmap:
            screen_image.filter_mipmap = GL.GL_LINEAR

        if self.camera is not None:
            self.program["camera"] = self.camera
            
        self.program["depth_map"] = self.depth_map
        self.program["view_pos_map"] = self.view_pos_map
        self.program["view_normal_map"] = self.view_normal_map
        self.program["screen_image"] = screen_image
        self._dynamic = (self.program["iTime"].location != -1 or \
                         self.program["iTimeDelta"].location != -1 or \
                         self.program["iFrameRate"].location != -1 or \
                         self.program["iFrame"].location != -1 or \
                         self.program["iDate"].location != -1)

        if self._dynamic:
            current_time = time.time()
            now = datetime.now()

            if self._start_time == 0:
                self._start_time = current_time
            if self._last_frame_time == 0:
                self._last_frame_time = current_time
            
            time_delta = current_time - self._last_frame_time
            fps = 60
            if time_delta > 0:
                fps = 1/time_delta
            t = current_time - self._start_time
            
            self.program["iTime"] = t
            self.program["iTimeDelta"] = time_delta
            self.program["iFrameRate"] = fps
            self.program["iFrame"] = self._frame_index
            self.program["iDate"] = glm.vec4(now.year, now.month, now.day, now.second + now.microsecond/1000)
            
            self._last_frame_time = current_time
            self._frame_index += 1
            
        self.program.draw_triangles(Frame.vertices, Frame.indices)

    def draw_to_active(self, screen_image:sampler2D)->None:
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.draw(screen_image)

    def apply(self, screen_image:sampler2D)->sampler2D:
        self.fbo.resize(screen_image.width, screen_image.height)
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            with self.fbo:
                self.draw(screen_image)

        return self.fbo.color_attachment(0)
    
    def __getitem__(self, name:str):
        if self._program is None:
            return self._uniforms[name]
        else:
            return self._program[name]
    
    def __setitem__(self, name:str, value):
        if self._program is None:
            self._uniforms[name] = value
        else:
            self.program[name] = value

    @property
    def should_update(self):
        if not self.enabled:
            return False
        
        return (self._should_update or self._dynamic)
    
    @should_update.setter
    def should_update(self, flag:bool):
        self._should_update = flag

    def need_pos_info(self)->bool:
        return (self.program["view_pos_map"].location >= 0 or \
                self.program["view_normal_map"].location >= 0 or \
                self.program["depth_map"].location >= 0)

    @staticmethod
    def __template(file_name):
        if not ShaderEffect.__template_content:
            ShaderEffect.__template_content = cat(ShaderEffect.__template_filename)
        
        rel_path = relative_path(file_name, GlassConfig.cache_folder)
        return ShaderEffect.__template_content.replace("{file_name}", rel_path)



# lextab.py. This file automatically created by PLY (version 3.11). Don't edit!
_tabversion   = '3.10'
_lextokens    = set(('CPP_AMPERSAND', 'CPP_ANDEQUAL', 'CPP_BAR', 'CPP_BSLASH', 'CPP_CHAR', 'CPP_COLON', 'CPP_COMMA', 'CPP_COMMENT1', 'CPP_COMMENT2', 'CPP_DEREFERENCE', 'CPP_DIVIDEEQUAL', 'CPP_DOT', 'CPP_DPOUND', 'CPP_DQUOTE', 'CPP_EQUAL', 'CPP_EQUALITY', 'CPP_EXCLAMATION', 'CPP_FLOAT', 'CPP_FSLASH', 'CPP_GREATER', 'CPP_GREATEREQUAL', 'CPP_HAT', 'CPP_ID', 'CPP_INEQUALITY', 'CPP_INTEGER', 'CPP_LBRACKET', 'CPP_LCURLY', 'CPP_LESS', 'CPP_LESSEQUAL', 'CPP_LINECONT', 'CPP_LOGICALAND', 'CPP_LOGICALOR', 'CPP_LPAREN', 'CPP_LSHIFT', 'CPP_LSHIFTEQUAL', 'CPP_MINUS', 'CPP_MINUSEQUAL', 'CPP_MINUSMINUS', 'CPP_MULTIPLYEQUAL', 'CPP_OREQUAL', 'CPP_PERCENT', 'CPP_PERCENTEQUAL', 'CPP_PLUS', 'CPP_PLUSEQUAL', 'CPP_PLUSPLUS', 'CPP_POUND', 'CPP_QUESTION', 'CPP_RBRACKET', 'CPP_RCURLY', 'CPP_RPAREN', 'CPP_RSHIFT', 'CPP_RSHIFTEQUAL', 'CPP_SEMICOLON', 'CPP_SQUOTE', 'CPP_STAR', 'CPP_STRING', 'CPP_TILDE', 'CPP_WS', 'CPP_XOREQUAL'))
_lexreflags   = 64
_lexliterals  = '+-*/%|&~^<>=!?()[]{}.,;:\\\'"'
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_CPP_WS>([ \\t]+|\\n))|(?P<t_CPP_LINECONT>\\\\[ \\t]*\\n)|(?P<t_CPP_INTEGER>(((((0x)|(0X))[0-9a-fA-F]+)|(\\d+))([uU][lL]|[lL][uU]|[uU]|[lL])?))|(?P<t_CPP_STRING>\\"([^\\\\\\n]|(\\\\(.|\\n)))*?\\")|(?P<t_CPP_CHAR>(L)?\\\'([^\\\\\\n]|(\\\\(.|\\n)))*?\\\')|(?P<t_CPP_COMMENT1>(/\\*(.|\\n)*?\\*/))|(?P<t_CPP_COMMENT2>(//[^\\n]*))|(?P<t_CPP_FLOAT>((\\d+)(\\.\\d+)(e(\\+|-)?(\\d+))?|(\\d+)e(\\+|-)?(\\d+))([lL]|[fF])?)|(?P<t_CPP_ID>[A-Za-z_][\\w_]*)|(?P<t_CPP_LOGICALOR>\\|\\|)|(?P<t_CPP_PLUSPLUS>\\+\\+)|(?P<t_CPP_DPOUND>\\#\\#)|(?P<t_CPP_LSHIFTEQUAL><<=)|(?P<t_CPP_OREQUAL>\\|=)|(?P<t_CPP_PLUSEQUAL>\\+=)|(?P<t_CPP_RSHIFTEQUAL>>>=)|(?P<t_CPP_MULTIPLYEQUAL>\\*=)|(?P<t_CPP_BAR>\\|)|(?P<t_CPP_DIVIDEEQUAL>/=)|(?P<t_CPP_POUND>\\#)|(?P<t_CPP_PERCENTEQUAL>%=)|(?P<t_CPP_DEREFERENCE>->)|(?P<t_CPP_RPAREN>\\))|(?P<t_CPP_ANDEQUAL>&=)|(?P<t_CPP_RBRACKET>\\])|(?P<t_CPP_LPAREN>\\()|(?P<t_CPP_RSHIFT>>>)|(?P<t_CPP_LESSEQUAL><=)|(?P<t_CPP_HAT>\\^)|(?P<t_CPP_LOGICALAND>&&)|(?P<t_CPP_EQUALITY>==)|(?P<t_CPP_GREATEREQUAL>>=)|(?P<t_CPP_BSLASH>\\\\)|(?P<t_CPP_MINUSEQUAL>-=)|(?P<t_CPP_DOT>\\.)|(?P<t_CPP_MINUSMINUS>--)|(?P<t_CPP_LBRACKET>\\[)|(?P<t_CPP_PLUS>\\+)|(?P<t_CPP_XOREQUAL>^=)|(?P<t_CPP_STAR>\\*)|(?P<t_CPP_QUESTION>\\?)|(?P<t_CPP_LSHIFT><<)|(?P<t_CPP_INEQUALITY>!=)|(?P<t_CPP_DQUOTE>")|(?P<t_CPP_MINUS>-)|(?P<t_CPP_RCURLY>})|(?P<t_CPP_GREATER>>)|(?P<t_CPP_LESS><)|(?P<t_CPP_SQUOTE>\')|(?P<t_CPP_EXCLAMATION>!)|(?P<t_CPP_LCURLY>{)|(?P<t_CPP_EQUAL>=)|(?P<t_CPP_FSLASH>/)|(?P<t_CPP_COLON>:)|(?P<t_CPP_AMPERSAND>&)|(?P<t_CPP_COMMA>,)|(?P<t_CPP_TILDE>~)|(?P<t_CPP_SEMICOLON>;)|(?P<t_CPP_PERCENT>%)', [None, ('t_CPP_WS', 'CPP_WS'), None, ('t_CPP_LINECONT', 'CPP_LINECONT'), ('t_CPP_INTEGER', 'CPP_INTEGER'), None, None, None, None, None, None, None, None, ('t_CPP_STRING', 'CPP_STRING'), None, None, None, ('t_CPP_CHAR', 'CPP_CHAR'), None, None, None, None, ('t_CPP_COMMENT1', 'CPP_COMMENT1'), None, None, ('t_CPP_COMMENT2', 'CPP_COMMENT2'), None, (None, 'CPP_FLOAT'), None, None, None, None, None, None, None, None, None, None, (None, 'CPP_ID'), (None, 'CPP_LOGICALOR'), (None, 'CPP_PLUSPLUS'), (None, 'CPP_DPOUND'), (None, 'CPP_LSHIFTEQUAL'), (None, 'CPP_OREQUAL'), (None, 'CPP_PLUSEQUAL'), (None, 'CPP_RSHIFTEQUAL'), (None, 'CPP_MULTIPLYEQUAL'), (None, 'CPP_BAR'), (None, 'CPP_DIVIDEEQUAL'), (None, 'CPP_POUND'), (None, 'CPP_PERCENTEQUAL'), (None, 'CPP_DEREFERENCE'), (None, 'CPP_RPAREN'), (None, 'CPP_ANDEQUAL'), (None, 'CPP_RBRACKET'), (None, 'CPP_LPAREN'), (None, 'CPP_RSHIFT'), (None, 'CPP_LESSEQUAL'), (None, 'CPP_HAT'), (None, 'CPP_LOGICALAND'), (None, 'CPP_EQUALITY'), (None, 'CPP_GREATEREQUAL'), (None, 'CPP_BSLASH'), (None, 'CPP_MINUSEQUAL'), (None, 'CPP_DOT'), (None, 'CPP_MINUSMINUS'), (None, 'CPP_LBRACKET'), (None, 'CPP_PLUS'), (None, 'CPP_XOREQUAL'), (None, 'CPP_STAR'), (None, 'CPP_QUESTION'), (None, 'CPP_LSHIFT'), (None, 'CPP_INEQUALITY'), (None, 'CPP_DQUOTE'), (None, 'CPP_MINUS'), (None, 'CPP_RCURLY'), (None, 'CPP_GREATER'), (None, 'CPP_LESS'), (None, 'CPP_SQUOTE'), (None, 'CPP_EXCLAMATION'), (None, 'CPP_LCURLY'), (None, 'CPP_EQUAL'), (None, 'CPP_FSLASH'), (None, 'CPP_COLON'), (None, 'CPP_AMPERSAND'), (None, 'CPP_COMMA'), (None, 'CPP_TILDE'), (None, 'CPP_SEMICOLON'), (None, 'CPP_PERCENT')])]}
_lexstateignore = {'INITIAL': ''}
_lexstateerrorf = {'INITIAL': 't_error'}
_lexstateeoff = {}



from .demo_transform import demo_transform
from .demo_material import demo_material

def demo(name:str):
    if name == "transform":
        demo_transform()
    elif name == "material":
        demo_material()


from functools import wraps
import ctypes
import re
import math
import glm
import os
import types
import cProfile
import hashlib
import pickle
import subprocess
import sys
import chardet
from _ctypes import PyObj_FromPtr

from .GlassConfig import GlassConfig

profiler = cProfile.Profile()

def run_exe(exe_path, *args):
    cmd = [exe_path, *args]
    result = subprocess.run(cmd, capture_output=True, text=True)
    stdout = result.stdout.strip()
    stderr = result.stderr.strip()
    return stdout, stderr

def cat(file_name):
    in_file = open(file_name, "r", encoding=sys.getdefaultencoding(), errors="ignore")
    content = in_file.read()
    in_file.close()
    return content

def bincat(file_name):
    in_file = open(file_name, "rb")
    content = in_file.read()
    in_file.close()
    return content

def uint64_to_uvec2(uint64_value):
    result = glm.uvec2()
    bytes_data = uint64_value.to_bytes(8, byteorder='little', signed=False)
    result.x = int.from_bytes(bytes_data[:4], byteorder='little', signed=False)
    result.y = int.from_bytes(bytes_data[4:], byteorder='little', signed=False)
    return result

def same_type(var1, var2):
    return (isinstance(var1, type(var2)) and isinstance(var2, type(var1)))

def checktype(func):
    if not GlassConfig.debug:
        return func
    
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not GlassConfig.debug:
            return func(*args, **kwargs)

        func_args = func.__code__.co_varnames[:func.__code__.co_argcount]
        annotation = func.__annotations__
        defaults = func.__defaults__
        len_func_args = len(func_args)

        len_defaults = 0
        if defaults is not None:
            len_defaults = len(defaults)
        defauts_start = len_func_args - len_defaults

        ext_kwargs = {}
        for i in range(defauts_start, len_func_args):
            ext_kwargs[func_args[i]] = defaults[i-defauts_start]
        ext_kwargs.update(kwargs)

        def change_annotation(arg_name):
            if annotation[arg_name] is None:
                annotation[arg_name] = (type(None),)

            if isinstance(annotation[arg_name], type):
                annotation[arg_name] = (annotation[arg_name], type(None))

            if float in annotation[arg_name] and \
                int not in annotation[arg_name]:
                annotation[arg_name] = (*annotation[arg_name], int)

            if None in annotation[arg_name]:
                type_set = set(annotation[arg_name])
                type_set.remove(None)
                type_set.add(type(None))
                annotation[arg_name] = tuple(type_set)

            if type(None) not in annotation[arg_name]:
                annotation[arg_name] = (*annotation[arg_name], type(None))

        for i in range(min(len(args), len_func_args)):
            arg_name = func_args[i]
            arg_value = args[i]

            if arg_name in annotation:
                if isinstance(annotation[arg_name], (type,tuple,type(None))):
                    change_annotation(arg_name)
                    if not isinstance(arg_value, annotation[arg_name]):
                        raise TypeError(func.__name__ + "() argument " + arg_name + " should be in type " + str(annotation[arg_name]) + \
                                        ", " + str(type(arg_value)) + " value passed.")
                elif isinstance(annotation[arg_name], (list,set)):
                    if arg_value not in annotation[arg_name]:
                        raise ValueError(func.__name__ + "() argument " + arg_name + " can only be in " + str(annotation[arg_name]) + \
                                         ", " + str(arg_value) + " passed.")

        for arg_name in ext_kwargs:
            arg_value = ext_kwargs[arg_name]
            if arg_name in annotation:
                if isinstance(annotation[arg_name], (type,tuple,type(None))):
                    change_annotation(arg_name)
                    if not isinstance(arg_value, annotation[arg_name]):
                        raise TypeError(func.__name__ + "() argument " + arg_name + " should be in type " + str(annotation[arg_name]) + \
                                        ", " + str(type(arg_value)) + " value passed.")
                elif isinstance(annotation[arg_name], (list,set)):
                    if arg_value not in annotation[arg_name]:
                        raise ValueError(func.__name__ + "() argument " + arg_name + " can only be in " + str(annotation[arg_name]) + \
                                         ", " + str(arg_value) + " passed.")

        return_value = func(*args, **kwargs)
        if "return" in annotation:
            if isinstance(annotation["return"], (type,tuple,type(None))):
                change_annotation("return")
                if not isinstance(return_value, annotation["return"]):
                    raise TypeError(func.__name__ + "() return value should be in type " + str(annotation["return"]) + \
                                    ", " + str(type(return_value)) + " value returned.")
            elif isinstance(annotation["return"], (list,set)):
                if return_value not in annotation["return"]:
                    raise ValueError(func.__name__ + "() return value can only be in " + str(annotation["return"]) + \
                                     ", " + str(return_value) + " returned.")

        return return_value

    return wrapper

def is_overridden(method):
    cls = method.__self__.__class__
    mro = cls.__mro__
    code = method.__code__
    method_name = method.__name__
    for i, base_cls in enumerate(mro):
        if method_name in base_cls.__dict__:
            base_method = base_cls.__dict__[method_name]
            if isinstance(base_method, types.FunctionType) and base_method.__code__ is not code:
                return True
    return False

def delete(method):
    @wraps(method)
    def wrapper(*args, **kwargs):
        self = args[0]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{method.__name__}'")
    return wrapper

def extname(filename):
    return os.path.splitext(filename)[1][1:].lower()

def has_valid(content):
    its = list(re.finditer(r"\S", content))
    return its.__bool__()

def rget_token(content, i):
    token = {}

    if i < 0 or i >= len(content):
        return "", -1

    i = rskip_space(content, i)
    if i < 0:
        return "", -1

    pos_start = i
    i = rskip_valid(content, i)
    pos_end = i

    token["span"] = (pos_end+1, pos_start+1)
    token["word"] = content[pos_end+1:pos_start+1]

    return token, i

def reg_rfind(content, pattern, pos_start, pos_end):
    str_index, comment_index = get_invalid_index(content)

    its = list(re.finditer(pattern, content[pos_start:pos_end]))
    if not its:
        return -1

    for it in reversed(its):
        if it.start() not in str_index and it.start() not in comment_index:
            return it.start()

    return it.start()

def rskip_space(content, i):
    if i < 0 or i >= len(content):
        return -1

    while True:
        if i < 0 or content[i] not in " \t\r.":
            break
        i -= 1

    return i

def rskip_valid(content, i):
    if i < 0 or i >= len(content):
        return -1

    if content[i] in "+-*/%!=<>,":
        i -= 1
        return i

    str_index, comment_index = get_invalid_index(content)
    while i >= 0:
        if i not in str_index and i not in comment_index:
            if content[i] in "\n\t+-*/%!=<>,([{":
                break

            if content[i] in " \t":
                j = rskip_space(content, i)
                if j == -1:
                    break

                if content[j] not in ")]}":
                    break
                else:
                    i = rfind_pair(content, j)

            elif content[i] in ")]}":
                i = rfind_pair(content, i)

        i -= 1

    return i

def find_pair(content, i):
    len_content = len(content)
    if i < 0 or i >= len_content or content[i] not in "([{":
        return -1

    start_pair = content[i]
    end_pair = None
    if content[i] == '(':
        end_pair = ')'
    elif content[i] == '[':
        end_pair = ']'
    elif content[i] == '{':
        end_pair = '}'

    n_start_pair = 0
    str_index, comment_index = get_invalid_index(content)
    while i < len_content:
        if i not in str_index and i not in comment_index:
            if content[i] == start_pair:
                n_start_pair += 1
            elif content[i] == end_pair:
                n_start_pair -= 1

            if n_start_pair == 0:
                break

        i += 1

    return i

def printable_size(content):
    len_content = len(content)
    KB = 1024
    MB = 1024 * KB
    GB = 1024 * MB
    TB = 1024 * GB
    PB = 1024 * TB

    if len_content < KB:
        return f"{len_content}B"
    elif len_content < MB:
        return f"{round(len_content/KB, 2)}KB"
    elif len_content < GB:
        return f"{round(len_content/MB, 2)}MB"
    elif len_content < TB:
        return f"{round(len_content/GB, 2)}GB"
    elif len_content < PB:
        return f"{round(len_content/TB, 2)}TB"
    else:
        return f"{round(len_content/PB, 2)}PB"

def is_text_file(file_path):
    if extname(file_path) in ["exr", "hdr"]:
        return False
    
    try:
        with open(file_path, 'rb') as f:
            result = chardet.detect(f.read())
            return (result['encoding'] is not None)
    except:
        return False

def is_url(line:str)->bool:
    pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    return bool(pattern.match(line))

def md5s(content):
    md5_hash = hashlib.md5()
    if isinstance(content, str):
        md5_hash.update(content.encode('utf-8'))
    elif isinstance(content, (bytes,bytearray)):
        md5_hash.update(content)
    else:
        md5_hash.update(pickle.dumps(content))

    return md5_hash.hexdigest()

def modify_time(file_name):
    if not os.path.isfile(file_name):
        return 0
    
    return os.path.getmtime(file_name)

def relative_path(file_name, start_path="."):
    try:
        return os.path.relpath(file_name, start_path).replace("\\", "/")
    except:
        return os.path.abspath(file_name).replace("\\", "/")

def printable_path(file_name):
    site_packages = os.path.abspath(os.path.dirname(os.path.abspath(__file__)) + "/../").replace("\\", "/")
    name1 = relative_path(file_name, site_packages)
    name2 = relative_path(file_name)
    name3 = os.path.abspath(file_name).replace("\\", "/")

    result = (name1 if len(name1) < len(name2) else name2)
    result = (result if len(result) < len(name3) else name3)
    return result

def save_var(var, file_path):
    with open(file_path, 'wb') as file:
        pickle.dump(var, file)

def load_var(file_path):
    with open(file_path, 'rb') as file:
        var = pickle.load(file)
    return var

def subscript(var, subscript_chain, feed_index:int=None):
    for operator, operant in subscript_chain:
        if operator == "getattr":
            var = getattr(var, operant)
        else:
            used_index = (operant if operant != "{0}" else feed_index)
            var = var[used_index]

    return var

def subscript_set(var, subscript_chain, value, feed_index:int=None, compare_before_set:bool=True):
    len_chain = len(subscript_chain)
    last_index = len_chain - 1
    for i in range(len_chain):
        item = subscript_chain[i]
        operator = item[0]
        operant = item[1]
        
        if operator == "getattr":
            old_value = getattr(var, operant)
            if i != last_index:
                var = old_value
            elif not compare_before_set or old_value != value:
                setattr(var, operant, value)
        else:
            used_index = (operant if operant != "{0}" else feed_index)
            old_value = var[used_index]
            if i != last_index:
                var = old_value
            elif not compare_before_set or old_value != value:
                var[used_index] = value

def get_subscript_chain(suffix):
    subscript_chain = []
    len_suffix = len(suffix)
    pos_start = 0
    while pos_start < len_suffix:
        start_char = suffix[pos_start]
        if start_char == "." or \
        'a' <= start_char <= 'z' or \
        'A' <= start_char <= 'Z':
            if start_char == ".":
                pos_start += 1

            pos_point = suffix.find(".", pos_start)
            if pos_point == -1:
                pos_point = len_suffix

            pos_braket = suffix.find("[", pos_start)
            if pos_braket == -1:
                pos_braket = len_suffix

            pos_end = min(pos_point, pos_braket)
            attr_name = suffix[pos_start:pos_end]
            subscript_chain.append(("getattr", attr_name))
            pos_start = pos_end
        elif start_char == "[":
            pos_end = find_pair(suffix, pos_start)
            if pos_end == -1:
                pos_end = len_suffix

            index = int(suffix[pos_start+1:pos_end])
            subscript_chain.append(("getitem", index))
            pos_start = pos_end + 1

    return subscript_chain

def rfind_pair(content, i):
    len_content = len(content)
    if i < 0 or i >= len_content or content[i] not in ")]}":
        return -1

    start_pair = content[i]
    end_pair = None
    if content[i] == ')':
        end_pair = '('
    elif content[i] == ']':
        end_pair = '['
    elif content[i] == '}':
        end_pair = '{'

    n_start_pair = 0
    str_index, comment_index = get_invalid_index(content)
    while i < len_content:
        if i not in str_index and i not in comment_index:
            if content[i] == start_pair:
                n_start_pair += 1
            elif content[i] == end_pair:
                n_start_pair -= 1

            if n_start_pair == 0:
                break

        i -= 1

    return i

__str_index_dict = {}
__comment_index_dict = {}
def get_invalid_index(content):
    global __str_index_dict
    if content in __str_index_dict:
        return __str_index_dict[content], __comment_index_dict[content]

    str_index = set()
    comment_index = set()
    in_comment = False
    in_str = False
    str_start_char = None
    last_is_slash = False
    i = 0
    len_content = len(content)
    while i < len_content:
        should_add = True
        if content[i] == "'" and not last_is_slash:
            if not in_str:
                should_add = False
                in_str = True
                str_start_char = "'"
            elif str_start_char == "'":
                in_str = False
                str_start_char = None
        elif content[i] == '"' and not last_is_slash:
            if not in_str:
                should_add = False
                in_str = True
                str_start_char = '"'
            elif str_start_char == '"':
                in_str = False
                str_start_char = None
        elif content[i] in "#\\" and not in_str:
            in_comment = True
        elif content[i] == "\n" and in_comment:
            in_comment = False
        
        if content[i] == '\\' and in_str:
            last_is_slash = (not last_is_slash)
        else:
            last_is_slash = False

        if should_add:
            if in_str:
                str_index.add(i)
            if in_comment:
                comment_index.add(i)

        i += 1

    __str_index_dict[content] = str_index
    __comment_index_dict[content] = comment_index

    return str_index, comment_index


def delete_python_comments(content):
    i = 0

    str_index, _ = get_invalid_index(content)
    len_content = len(content)

    while i < len_content:
        if i not in str_index:
            if content[i] == "#":
                pos_endl = content.find("\n", i)
                if pos_endl == -1:
                    pos_endl = len_content
                content = content[:i] + content[pos_endl:]
            elif content[i] == '\\':
                pos_endl = content.find("\n", i)
                if pos_endl == -1:
                    pos_endl = len_content
                else:
                    pos_endl += 1

                content = content[:i] + content[pos_endl:]

        i += 1

    return content

def split_var_str(var_str):
    pos = len(var_str)-1
    if var_str[pos] in ")]}":
        while pos >= 0 and var_str[pos] in ")]}":
            pos = rfind_pair(var_str, pos)
            pos -= 1
            pos = rskip_space(var_str, pos)
        return var_str[:pos+1], var_str[pos+1:]
    else:
        pos = var_str.rfind(".")
        if pos != -1:
            return var_str[:pos], var_str[pos:]
        else:
            return "", ""

def di(id_var):
    return PyObj_FromPtr(id_var)

def LP_LP_c_char(str_list):
    lp_c_char = ctypes.POINTER(ctypes.c_char)
    len_str_list = len(str_list)
    result = (lp_c_char*len_str_list)()
    for i in range(len_str_list):
        result[i] = ctypes.create_string_buffer(str_list[i].encode("utf-8"))
    return result

def capacity_of(length):
    if length <= 0:
        return 2
    else:
        return int(math.pow(2, math.floor(1 + math.log2(length))))

@checktype
def quat_to_mat4(q:glm.quat):
    w = q.w
    x = q.x
    y = q.y
    z = q.z

    return glm.mat4(1-2*(y**2 + z**2), 2*(x*y + w*z), 2*(x*z - w*y), 0,
                     2*(x*y - w*z), 1-2*(x**2 + z**2), 2*(y*z + w*x), 0,
                     2*(x*z + w*y), 2*(y*z - w*x), 1-2*(x**2 + y**2), 0,
                     0, 0, 0, 1)

@checktype
def quat_to_mat3(q:glm.quat):
    w = q.w
    x = q.x
    y = q.y
    z = q.z

    return glm.mat3(1-2*(y**2 + z**2), 2*(x*y + w*z), 2*(x*z - w*y),
                     2*(x*y - w*z), 1-2*(x**2 + z**2), 2*(y*z + w*x),
                     2*(x*z + w*y), 2*(y*z - w*x), 1-2*(x**2 + y**2))

def scale_to_mat4(s:glm.vec3):
    return glm.mat4(s.x, 0  , 0  , 0,
                    0  , s.y, 0  , 0,
                    0  , 0  , s.z, 0,
                    0  , 0  , 0  , 1)

def scale_to_mat3(s:glm.vec3):
    return glm.mat3(s.x, 0  , 0  ,
                    0  , s.y, 0  ,
                    0  , 0  , s.z)

def translate_to_mat4(t:glm.vec3):
    return glm.mat4(1  , 0  , 0  , 0,
                    0  , 1  , 0  , 0,
                    0  , 0  , 1  , 0,
                    t.x, t.y, t.z, 1)

def defines_key(*args):
    shader_type_defines = \
    {
        "VERTEX_SHADER",
        "TESS_CONTROL_SHADER",
        "TESS_EVALUATION_SHADER",
        "GEOMETRY_SHADER",
        "FRAGMENT_SHADER",
        "COMPUTE_SHADER"
    }

    defines = {}
    for arg in args:
        defines.update(arg)

    keys = list(defines.keys())
    keys.sort()
    defines_str = ""
    for key in keys:
        if key in shader_type_defines:
            continue

        part = key
        if defines[key] is not None:
            part += f"={defines[key]}"
        defines_str += (":" + part)

    return defines_str


from OpenGL import GL
import OpenGL.GL.ARB.gpu_shader_int64 as gsi64
import glm
import copy
from enum import Enum

from .utils import checktype, get_subscript_chain, uint64_to_uvec2, di
from .sampler2D import sampler2D
from .image2D import image2D
from .FBOAttachment import FBOAttachment
from .usampler2D import usampler2D
from .isampler2D import isampler2D
from .uimage2D import uimage2D
from .iimage2D import iimage2D
from .sampler2DMS import sampler2DMS
from .isampler2DMS import isampler2DMS
from .usampler2DMS import usampler2DMS
from .samplerCube import samplerCube
from .sampler2DArray import sampler2DArray
from .ACBO import ACBO
from .GlassConfig import GlassConfig
from .utils import subscript

class Uniform:
    
    _bound_vars = {}
    # {
    #     "<id_normal_var>":
    #     {
    #         "<atom_name>":
    #         {
    #             "suffix": "<suffix>"
    #             "uniforms": set()
    #         }
    #     }
    # }

    _set_atom_map = {}

    class Variable:
        _all_attrs = {"__init__", "__getitem__", "__setitem__", "__getattr__", "__setattr__",
                      "_uniform_id", "_name", "_bound_var",
                      "bind", "unbind", "location"}

        def __init__(self, uniform_id, name):
            self._uniform_id = uniform_id
            self._name = name
            self._bound_var = None

        @property
        def uniform(self):
            return di(self._uniform_id)

        def __del__(self):
            self.unbind()

        def __hash__(self):
            return id(self)
        
        def __eq__(self, other):
            return (id(self) == id(other))

        @property
        def location(self):
            program = self.uniform.program
            if "location" not in program._uniform_map[self._name]:
                program.use()
                location = GL.glGetUniformLocation(program._id, self._name)
                program._uniform_map[self._name]["location"] = location
                return location
            else:
                return program._uniform_map[self._name]["location"]

        def bind(self, var):
            if var is self._bound_var:
                return

            if var is None:
                self.unbind()
                return

            if isinstance(var, (bool, int, float)):
                raise ValueError("cannot bind to " + var.__class__.__name__ + " value")

            self.unbind()

            id_var = id(var)
            if id_var not in Uniform._bound_vars:
                Uniform._bound_vars[id_var] = {}

            len_name = len(self._name)
            for atom in self.uniform.program._uniform_map[self._name]["atoms"]:
                atom_name = atom["name"]
                atom_suffix = atom["name"][len_name:]
                subscript_chain = get_subscript_chain(atom_suffix)
                subscript(var, subscript_chain)
                if atom_name not in Uniform._bound_vars[id_var]:
                    Uniform._bound_vars[id_var][atom_name] = {}

                atom_info = Uniform._bound_vars[id_var][atom_name]
                atom_info["suffix"] = atom_suffix
                atom_info["subscript_chain"] = subscript_chain
                if "uniforms" not in atom_info:
                    atom_info["uniforms"] = set()
                atom_info["uniforms"].add(self._uniform_id)

            self._bound_var = var

        def unbind(self):
            if self._bound_var is None:
                return

            id_var = id(self._bound_var)
            if id_var in Uniform._bound_vars:
                var_info = Uniform._bound_vars[id_var]
                should_remove_atoms = []
                for atom_name, info in var_info.items():
                    if self._uniform_id in info["uniforms"]:
                        info["uniforms"].remove(self._uniform_id)
                    if not info["uniforms"]:
                        should_remove_atoms.append(atom_name)
                for atom_name in should_remove_atoms:
                    del var_info[atom_name]
                if not var_info:
                    del Uniform._bound_vars[id_var]
            
            self._bound_var = None

        def __contains__(self, name:(str, int)):
            full_name = self._name
            if isinstance(name, str):
                full_name += ("." + name)
            elif isinstance(name, int):
                full_name += "[" + str(name) + "]"

            return (full_name in self.uniform.program._uniform_map)

        def __getitem__(self, name:(str, int)):
            full_name = self._name
            if isinstance(name, str):
                full_name += ("." + name)
            elif isinstance(name, int):
                full_name += "[" + str(name) + "]"

            uniform = self.uniform
            program = uniform.program
            if GlassConfig.debug and full_name not in program._uniform_map:
                error_message = "uniform variable '" + full_name + "' is not defined in following files:\n"
                error_message += "\n".join(program.related_files)
                raise NameError(error_message)
            
            if full_name not in uniform._uniform_var_map:
                uniform._uniform_var_map[full_name] = Uniform.Variable(self._uniform_id, full_name)

            return uniform._uniform_var_map[full_name]

        @checktype
        def __setitem__(self, name:(str, int), value):
            full_name = self._name
            if isinstance(name, str):
                full_name += ("." + name)
            elif isinstance(name, int):
                full_name += "[" + str(name) + "]"

            uniform = self.uniform
            program = uniform.program
            if GlassConfig.debug and full_name not in program._uniform_map:
                error_message = "uniform variable '" + full_name + "' is not defined in following files:\n"
                error_message += "\n".join(program.related_files)
                raise NameError(error_message)

            uniform[full_name] = value

        def __getattr__(self, name:str):
            if name in Uniform.Variable._all_attrs:
                return super().__getattr__(name)

            return self.__getitem__(name)

        def __setattr__(self, name:str, value):
            if name in Uniform.Variable._all_attrs:
                return super().__setattr__(name, value)

            self.__setitem__(name, value)

    def __init__(self, shader_program):
        self._program_id = id(shader_program)
        self._atoms_to_update = {}
        self._uniform_var_map = {}
        self._atom_value_map = {}
        self._texture_value_map = {}
        self._current_atom_name = ""

    @property
    def program(self):
        return di(self._program_id)

    def __getitem__(self, name:str):
        program = self.program
        if GlassConfig.debug and name not in program._uniform_map:
            error_message = f"uniform variable '{name}' is not defined in following files:\n"
            error_message += "\n".join(program.related_files)
            raise NameError(error_message)

        if name not in self._uniform_var_map:
            self._uniform_var_map[name] = Uniform.Variable(id(self), name)

        return self._uniform_var_map[name]

    def __setitem__(self, name:str, value):
        program = self.program
        if GlassConfig.debug and name not in program._uniform_map:
            error_message = f"uniform variable '{name}' is not defined in following files:\n"
            error_message += "\n".join(program.related_files)
            raise NameError(error_message)

        for atom in program._uniform_map[name]["atoms"]:
            atom_value = subscript(value, atom["subscript_chain"])
            self._set_atom(atom["name"], atom_value)

    @checktype
    def __contains__(self, name:str):
        return (name in self.program._uniform_map)

    @staticmethod
    def _copy(value):
        type_str = type(value).__name__
        if value is None or "sampler" in type_str or "image" in type_str:
            return value
        else:
            return copy.copy(value)

    def _set_atom(self, name:str, value):
        if name in self._atom_value_map and self._atom_value_map[name] == value:
            if isinstance(value, FBOAttachment):
                value.bind()
            return
        
        if GL.glGetUniformLocation:
            program = self.program
            program.use()

            uniform_info = program._uniform_map[name]
            uniform_type = uniform_info["type"]
            if uniform_type != "atomic_uint":
                location = -1
                if "location" not in uniform_info:
                    location = GL.glGetUniformLocation(program._id, name)
                    uniform_info["location"] = location
                else:
                    location = uniform_info["location"]

                if location == -1:
                    return
                
                self._current_atom_name = name
                func = Uniform._set_atom_func(uniform_type)
                func(self, location, value)
                self._atom_value_map[name] = Uniform._copy(value)
            else:
                binding_point = uniform_info["binding_point"]
                offset = uniform_info["offset"]
                ACBO.set(binding=binding_point, offset=offset, value=value)
        else:
            self._atoms_to_update[name] = Uniform._copy(value)

    @staticmethod
    def _set_atom_func(atom_type):
        if not Uniform._set_atom_map:
            Uniform._set_atom_map = \
            {
                "bool": Uniform._set_bool, "int": Uniform._set_int, "uint": Uniform._set_uint,
                "uint64_t": Uniform._set_uint64_t, "float": Uniform._set_float, "double": Uniform._set_double,
                "bvec2": Uniform._set_bvec2, "bvec3": Uniform._set_bvec3, "bvec4": Uniform._set_bvec4,
                "ivec2": Uniform._set_ivec2, "ivec3": Uniform._set_ivec3, "ivec4": Uniform._set_ivec4,
                "uvec2": Uniform._set_uvec2, "uvec3": Uniform._set_uvec3, "uvec4": Uniform._set_uvec4,
                "vec2": Uniform._set_vec2, "vec3": Uniform._set_vec3, "vec4": Uniform._set_vec4,
                "dvec2": Uniform._set_dvec2, "dvec3": Uniform._set_dvec3, "dvec4": Uniform._set_dvec4,
                "mat2": Uniform._set_mat2, "mat3": Uniform._set_mat3, "mat4": Uniform._set_mat4,
                "dmat2": Uniform._set_dmat2, "dmat3": Uniform._set_dmat3, "dmat4": Uniform._set_dmat4,
                "mat2x2": Uniform._set_mat2x2, "mat2x3": Uniform._set_mat2x3, "mat2x4": Uniform._set_mat2x4,
                "mat3x2": Uniform._set_mat3x2, "mat3x3": Uniform._set_mat3x3, "mat3x4": Uniform._set_mat3x4,
                "mat4x2": Uniform._set_mat4x2, "mat4x3": Uniform._set_mat4x3, "mat4x4": Uniform._set_mat4x4,
                "dmat2x2": Uniform._set_dmat2x2, "dmat2x3": Uniform._set_dmat2x3, "dmat2x4": Uniform._set_dmat2x4,
                "dmat3x2": Uniform._set_dmat3x2, "dmat3x3": Uniform._set_dmat3x3, "dmat3x4": Uniform._set_dmat3x4,
                "dmat4x2": Uniform._set_dmat4x2, "dmat4x3": Uniform._set_dmat4x3, "dmat4x4": Uniform._set_dmat4x4,
            
                "sampler2D": Uniform._set_sampler2D, "isampler2D": Uniform._set_isampler2D, "usampler2D": Uniform._set_usampler2D,
                "image2D": Uniform._set_image2D, "iimage2D": Uniform._set_iimage2D, "uimage2D": Uniform._set_uimage2D,
                "sampler2DMS": Uniform._set_sampler2DMS, "isampler2DMS": Uniform._set_isampler2DMS, "usampler2DMS": Uniform._set_usampler2DMS,
                "samplerCube": Uniform._set_samplerCube, "sampler2DArray": Uniform._set_sampler2DArray
            }
        return Uniform._set_atom_map[atom_type]

    def _set_bool(self, location:int, value):
        if not isinstance(value, int):
            value = int(value)

        GL.glUniform1i(location, value)

    def _set_int(self, location:int, value):
        if isinstance(value, Enum):
            value = int(value.value)

        if not isinstance(value, int):
            value = int(value)

        GL.glUniform1i(location, value)

    def _set_uint(self, location:int, value):
        if isinstance(value, Enum):
            value = int(value.value)

        if not isinstance(value, int):
            value = int(value)
            
        GL.glUniform1ui(location, value)

    def _set_uint64_t(self, location:int, value):
        if not isinstance(value, int):
            value = int(value)

        gsi64.glUniform1ui64ARB(location, value)

    def _set_float(self, location:int, value):
        if not isinstance(value, float):
            value = float(value)

        GL.glUniform1f(location, value)

    def _set_double(self, location:int, value):
        if not isinstance(value, float):
            value = float(value)

        GL.glUniform1d(location, value)

    def _set_bvec2(self, location:int, value:glm.bvec2):
        if not isinstance(value, glm.bvec2):
            value = glm.bvec2(value)

        GL.glUniform2i(location, int(value.x), int(value.y))

    def _set_bvec3(self, location:int, value:glm.bvec3):
        if not isinstance(value, glm.bvec3):
            value = glm.bvec3(value)

        GL.glUniform3i(location, int(value.x), int(value.y), int(value.z))

    def _set_bvec4(self, location:int, value:glm.bvec4):
        if not isinstance(value, glm.bvec4):
            value = glm.bvec4(value)

        GL.glUniform4i(location, int(value.x), int(value.y), int(value.z), int(value.w))

    def _set_ivec2(self, location:int, value:glm.ivec2):
        if not isinstance(value, glm.ivec2):
            value = glm.ivec2(value)

        GL.glUniform2i(location, value.x, value.y)
        
    def _set_ivec3(self, location:int, value:glm.ivec3):
        if not isinstance(value, glm.ivec3):
            value = glm.ivec3(value)

        GL.glUniform3i(location, value.x, value.y, value.z)
        
    def _set_ivec4(self, location:int, value:glm.ivec4):
        if not isinstance(value, glm.ivec4):
            value = glm.ivec4(value)

        GL.glUniform4i(location, value.x, value.y, value.z, value.w)
            
    def _set_uvec2(self, location:int, value:(glm.uvec2,int)):
        if isinstance(value, glm.uvec2):
            GL.glUniform2ui(location, value.x, value.y)
        else:
            if not isinstance(value, int):
                value = int(value)

            used_value = uint64_to_uvec2(value)
            GL.glUniform2ui(location, used_value.x, used_value.y)
            
    def _set_uvec3(self, location:int, value:glm.uvec3):
        if not isinstance(value, glm.uvec3):
            value = glm.uvec3(value)

        GL.glUniform3ui(location, value.x, value.y, value.z)
            
    def _set_uvec4(self, location:int, value:glm.uvec4):
        if not isinstance(value, glm.uvec4):
            value = glm.uvec4(value)

        GL.glUniform4ui(location, value.x, value.y, value.z, value.w)
            
    def _set_vec2(self, location:int, value:glm.vec2):
        if not isinstance(value, glm.vec2):
            value = glm.vec2(value)

        GL.glUniform2f(location, value.x, value.y)
        
    def _set_vec3(self, location:int, value:glm.vec3):
        if not isinstance(value, glm.vec3):
            value = glm.vec3(value)

        GL.glUniform3f(location, value.x, value.y, value.z)
            
    def _set_vec4(self, location:int, value:glm.vec4):
        if not isinstance(value, glm.vec4):
            value = glm.vec4(value)

        GL.glUniform4f(location, value.x, value.y, value.z, value.w)
        
    def _set_dvec2(self, location:int, value:glm.dvec2):
        if not isinstance(value, glm.dvec2):
            value = glm.dvec2(value)

        GL.glUniform2d(location, value.x, value.y)
        
    def _set_dvec3(self, location:int, value:glm.dvec3):
        if not isinstance(value, glm.dvec3):
            value = glm.dvec3(value)

        GL.glUniform3d(location, value.x, value.y, value.z)
            
    def _set_dvec4(self, location:int, value:glm.dvec4):
        if not isinstance(value, glm.dvec4):
            value = glm.dvec4(value)

        GL.glUniform4d(location, value.x, value.y, value.z, value.w)
            
    def _set_mat2(self, location:int, value:glm.mat2):
        if not isinstance(value, glm.mat2):
            value = glm.mat2(value)

        GL.glUniformMatrix2fv(location, 1, False, glm.value_ptr(value))
            
    def _set_mat3x2(self, location:int, value:glm.mat3x2):
        if not isinstance(value, glm.mat3x2):
            value = glm.mat3x2(value)

        GL.glUniformMatrix3x2fv(location, 1, False, glm.value_ptr(value))
        
    def _set_mat4x2(self, location:int, value:glm.mat4x2):
        if not isinstance(value, glm.mat4x2):
            value = glm.mat4x2(value)

        GL.glUniformMatrix4x2fv(location, 1, False, glm.value_ptr(value))
        
    def _set_mat2x3(self, location:int, value:glm.mat2x3):
        if not isinstance(value, glm.mat2x3):
            value = glm.mat2x3(value)

        GL.glUniformMatrix2x3fv(location, 1, False, glm.value_ptr(value))
            
    def _set_mat3(self, location:int, value:glm.mat3x3):
        if not isinstance(value, glm.mat3x3):
            value = glm.mat3x3(value)

        GL.glUniformMatrix3fv(location, 1, False, glm.value_ptr(value))
            
    def _set_mat4x3(self, location:int, value:glm.mat4x3):
        if not isinstance(value, glm.mat4x3):
            value = glm.mat4x3(value)

        GL.glUniformMatrix4x3fv(location, 1, False, glm.value_ptr(value))
            
    def _set_mat2x4(self, location:int, value:glm.mat2x4):
        if not isinstance(value, glm.mat2x4):
            value = glm.mat2x4(value)

        GL.glUniformMatrix2x4fv(location, 1, False, glm.value_ptr(value))
            
    def _set_mat3x4(self, location:int, value:glm.mat3x4):
        if not isinstance(value, glm.mat3x4):
            value = glm.mat3x4(value)

        GL.glUniformMatrix3x4fv(location, 1, False, glm.value_ptr(value))
            
    def _set_mat4(self, location:int, value:glm.mat4x4):
        if not isinstance(value, glm.mat4x4):
            value = glm.mat4x4(value)

        GL.glUniformMatrix4fv(location, 1, False, glm.value_ptr(value))
        
    def _set_mat2x2(self, location:int, value:glm.mat2x2):
        if not isinstance(value, glm.mat2x2):
            value = glm.mat2x2(value)

        GL.glUniformMatrix2fv(location, 1, False, glm.value_ptr(value))
        
    def _set_mat3x3(self, location:int, value:glm.mat3x3):
        if not isinstance(value, glm.mat3x3):
            value = glm.mat3x3(value)

        GL.glUniformMatrix3fv(location, 1, False, glm.value_ptr(value))
        
    def _set_mat4x4(self, location:int, value:glm.mat4x4):
        if not isinstance(value, glm.mat4x4):
            value = glm.mat4x4(value)

        GL.glUniformMatrix4fv(location, 1, False, glm.value_ptr(value))
        
    def _set_dmat2(self, location:int, value:glm.dmat2):
        if not isinstance(value, glm.dmat2):
            value = glm.dmat2(value)

        GL.glUniformMatrix2dv(location, 1, False, glm.value_ptr(value))
            
    def _set_dmat3x2(self, location:int, value:glm.dmat3x2):
        if not isinstance(value, glm.dmat3x2):
            value = glm.dmat3x2(value)

        GL.glUniformMatrix3x2dv(location, 1, False, glm.value_ptr(value))
        
    def _set_dmat4x2(self, location:int, value:glm.dmat4x2):
        if not isinstance(value, glm.dmat4x2):
            value = glm.dmat4x2(value)

        GL.glUniformMatrix4x2dv(location, 1, False, glm.value_ptr(value))
        
    def _set_dmat2x3(self, location:int, value:glm.dmat2x3):
        if not isinstance(value, glm.dmat2x3):
            value = glm.dmat2x3(value)

        GL.glUniformMatrix2x3dv(location, 1, False, glm.value_ptr(value))
            
    def _set_dmat3(self, location:int, value:glm.dmat3x3):
        if not isinstance(value, glm.dmat3x3):
            value = glm.dmat3x3(value)

        GL.glUniformMatrix3dv(location, 1, False, glm.value_ptr(value))
            
    def _set_dmat4x3(self, location:int, value:glm.dmat4x3):
        if not isinstance(value, glm.dmat4x3):
            value = glm.dmat4x3(value)

        GL.glUniformMatrix4x3dv(location, 1, False, glm.value_ptr(value))
            
    def _set_dmat2x4(self, location:int, value:glm.dmat2x4):
        if not isinstance(value, glm.dmat2x4):
            value = glm.dmat2x4(value)

        GL.glUniformMatrix2x4dv(location, 1, False, glm.value_ptr(value))
            
    def _set_dmat3x4(self, location:int, value:glm.dmat3x4):
        if not isinstance(value, glm.dmat3x4):
            value = glm.dmat3x4(value)

        GL.glUniformMatrix3x4dv(location, 1, False, glm.value_ptr(value))
            
    def _set_dmat4(self, location:int, value:glm.dmat4):
        if not isinstance(value, glm.dmat4):
            value = glm.dmat4(value)

        GL.glUniformMatrix4dv(location, 1, False, glm.value_ptr(value))
        
    def _set_dmat2x2(self, location:int, value:glm.dmat2x2):
        if not isinstance(value, glm.dmat2x2):
            value = glm.dmat2x2(value)

        GL.glUniformMatrix2dv(location, 1, False, glm.value_ptr(value))
        
    def _set_dmat3x3(self, location:int, value:glm.dmat3x3):
        if not isinstance(value, glm.dmat3x3):
            value = glm.dmat3x3(value)

        GL.glUniformMatrix3dv(location, 1, False, glm.value_ptr(value))
        
    def _set_dmat4x4(self, location:int, value:glm.dmat4x4):
        if not isinstance(value, glm.dmat4x4):
            value = glm.dmat4x4(value)

        GL.glUniformMatrix4dv(location, 1, False, glm.value_ptr(value))
        
    @checktype
    def _set_sampler2D(self, location:int, value:(sampler2D,str), sampler_type:[sampler2D,isampler2D,usampler2D]=sampler2D):
        if isinstance(value, str):
            value = sampler_type.load(value)

        if value is not None:
            value.bind()

        program = self.program
        program._sampler_map[self._current_atom_name]["location"] = location
        program._sampler_map[self._current_atom_name]["sampler"] = value
        
    @checktype
    def _set_isampler2D(self, location:int, value:(isampler2D,str)):
        self._set_sampler2D(location, value, isampler2D)
        
    @checktype
    def _set_usampler2D(self, location:int, value:(usampler2D,str)):
        self._set_sampler2D(location, value, usampler2D)
        
    @checktype
    def _set_sampler2DMS(self, location:int, value:sampler2DMS):
        if value is not None:
            value.bind()

        program = self.program
        program._sampler_map[self._current_atom_name]["location"] = location
        program._sampler_map[self._current_atom_name]["sampler"] = value
        
    @checktype
    def _set_isampler2DMS(self, location:int, value:isampler2DMS):
        self._set_sampler2DMS(location, value)
        
    @checktype
    def _set_usampler2DMS(self, location:int, value:usampler2DMS):
        self._set_sampler2DMS(location, value)
        
    @checktype
    def _set_sampler2DArray(self, location:int, value:sampler2DArray):
        if value is not None:
            value.bind()

        program = self.program
        program._sampler_map[self._current_atom_name]["location"] = location
        program._sampler_map[self._current_atom_name]["sampler"] = value

    def __set_general_image2D(self, location:int, value:(image2D,iimage2D,uimage2D,str), image_type:[image2D,iimage2D,uimage2D]=image2D):
        if isinstance(value, str):
            value = image_type.load(value, internal_format=self._current_internal_format)

        access = GL.GL_READ_WRITE
        program = self.program
        memory_qualifiers = program._uniform_map[self._current_atom_name]["memory_qualifiers"]
        internal_format = program._uniform_map[self._current_atom_name]["internal_format"]
        if value is not None:
            if GlassConfig.debug and internal_format != value.internal_format:
                raise ValueError(f"uniform image2D {self._current_atom_name} need format {internal_format}, {value.internal_format} were given")

            access = GL.GL_READ_WRITE
            if "readonly" in memory_qualifiers:
                access = GL.GL_READ_ONLY
            elif "writeonly" in memory_qualifiers:
                access = GL.GL_WRITE_ONLY
            
            value.bind()
            
        program._sampler_map[self._current_atom_name]["location"] = location
        program._sampler_map[self._current_atom_name]["access"] = access
        program._sampler_map[self._current_atom_name]["sampler"] = value
        
    @checktype
    def _set_iimage2D(self, location:int, value:(iimage2D,str)):
        self.__set_general_image2D(location, value, image_type=iimage2D)
        
    @checktype
    def _set_uimage2D(self, location:int, value:(uimage2D,str)):
        self.__set_general_image2D(location, value, image_type=uimage2D)
        
    @checktype
    def _set_image2D(self, location:int, value:(image2D,str)):
        self.__set_general_image2D(location, value, image_type=image2D)
        
    @checktype
    def _set_samplerCube(self, location:int, value:samplerCube):
        if value is not None and not value.is_completed:
            raise RuntimeError("not completed samplerCube")

        if value is not None:
            value.bind()

        program = self.program
        program._sampler_map[self._current_atom_name]["sampler"] = value
        program._sampler_map[self._current_atom_name]["location"] = location



from .Mesh import Mesh
from .Camera import Camera
from .GlassEngineConfig import GlassEngineConfig

from glass.utils import checktype
from glass import Vertex, sampler2D, ShaderProgram

import glm
import math
import numpy as np
from OpenGL import GL
import os

class SkyDome(Mesh):

    @checktype
    def __init__(self, n_lon_divide:int=100, n_lat_divide:int=50, name:str=""):
        Mesh.__init__(self, name=name, block=True)
        self.__n_lon_divide = n_lon_divide
        self.__n_lat_divide = n_lat_divide
        self.__skydome_map = None
        self.__program = None

        self.render_hints.depth_func = "<="
        self.render_hints.cull_face = GL.GL_FRONT
        self.should_add_color = False
        self.start_building()

    def build(self):
        self.is_closed = True
        self.self_calculated_normal = True
        self.x_min, self.x_max = -1, 1
        self.y_min, self.y_max = -1, 1
        self.z_min, self.z_max = -1, 1

        vertices = self.vertices
        indices = self.indices
        n_lon_divide = self.__n_lon_divide
        n_lat_divide = self.__n_lat_divide

        i_vertex = 0
        i_index = 0

        for i in range(n_lon_divide):
            theta = math.pi*(2*i/(n_lon_divide - 1) - 0.5)
            s = 1 - i/(n_lon_divide - 1)
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)

            for j in range(n_lat_divide):
                phi = math.pi*(j/(n_lat_divide - 1) - 0.5)
                t = j/(n_lat_divide - 1)
                cos_phi = math.cos(phi)
                sin_phi = math.sin(phi)

                vertex = Vertex()

                vertex.position = glm.vec3(cos_phi*cos_theta, cos_phi*sin_theta, sin_phi)
                vertex.tex_coord = glm.vec2(s, t)

                vertices[i_vertex] = vertex
                i_vertex += 1

                if i > 0 and j > 0:
                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex-1
                    triangle[1] = i_vertex-1-n_lat_divide
                    triangle[2] = i_vertex-1-n_lat_divide-1
                    indices[i_index] = triangle
                    i_index += 1

                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex-1
                    triangle[1] = i_vertex-1-n_lat_divide-1
                    triangle[2] = i_vertex-1-1
                    indices[i_index] = triangle
                    i_index += 1
        
        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def n_lat_divide(self):
        return self.__n_lat_divide
    
    @n_lat_divide.setter
    @Mesh.param_setter
    def n_lat(self, n_lat_divide:int):
        self.__n_lat_divide = n_lat_divide

    @property
    def n_lon_divide(self):
        return self.__n_lon_divide
    
    @n_lon_divide.setter
    @Mesh.param_setter
    def n_lon(self, n_lon_divide:int):
        self.__n_lon_divide = n_lon_divide

    @property
    def skydome_map(self):
        return self.__skydome_map
    
    @skydome_map.setter
    @checktype
    def skydome_map(self, skydome_map:(sampler2D,np.ndarray,str)):
        if isinstance(skydome_map, (np.ndarray,str)):
            if self.__skydome_map is None:
                self.__skydome_map = sampler2D()
            self.__skydome_map.image = skydome_map
        else:
            self.__skydome_map = skydome_map

    @property
    def is_completed(self):
        return (self.__skydome_map is not None and self.__skydome_map.is_completed)
    
    def __getitem__(self, name:str):
        return self.__skydome_map[name]
    
    def __setitem__(self, name:str, value):
        self.__skydome_map[name] = value

    @property
    def program(self):
        if self.__program is None:
            self.__program = ShaderProgram()
            self_folder = os.path.dirname(os.path.abspath(__file__))
            self.__program.compile(self_folder + "/glsl/Pipelines/skydome/skydome.vs")
            self.__program.compile(self_folder + "/glsl/Pipelines/skydome/skydome.fs")

        return self.__program
    
    def draw(self, camera:Camera):
        scene = camera.scene
        self.program["camera"] = camera
        self.program["skydome_map"] = self.skydome_map
        self.program["sky_distance"] = scene.background.distance

        if GlassEngineConfig["USE_FOG"]:
            self.program["fog"] = scene.fog
            
        Mesh.draw(self, self.program)


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel, QSpinBox
from PyQt6.QtCore import Qt, pyqtSignal

class IntInput(QWidget):

    value_changed = pyqtSignal(int)

    def __init__(self, prompt:str, range, value:int=None, parent:QWidget=None):
        QWidget.__init__(self, parent)

        hlayout = QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        label = QLabel(prompt, self)
        hlayout.addWidget(label)

        self.__spinbox = QSpinBox()
        self.__spinbox.setMinimum(range[0])
        self.__spinbox.setMaximum(range[1])
        if value is not None:
            self.__spinbox.setValue(value)
        hlayout.addWidget(self.__spinbox)

        self.__spinbox.valueChanged.connect(self.value_changed)

    @property
    def value(self):
        return self.__spinbox.value()
    
    @value.setter
    def value(self, value:float):
        self.__spinbox.setValue(value)




from .SameTypeList import SameTypeList
from .GLInfo import GLInfo
from .helper import sizeof
from .VBO import VBO
from .Increment import Increment

from OpenGL import GL
import numpy as np
import glm

class AttrList(SameTypeList):

    def __init__(self, _list:(list,np.ndarray)=None, draw_type:GLInfo.draw_types=GL.GL_STATIC_DRAW, dtype:GLInfo.attr_types=None):
        SameTypeList.__init__(self, _list, dtype)

        self._draw_type = draw_type
        self._vbo = VBO()

        self.stride = 0
        self.is_new_vbo = False
    
    @property
    def draw_type(self)->GLInfo.draw_types:
        return self._draw_type
    
    @draw_type.setter
    def draw_type(self, draw_type:GLInfo.draw_types):
        self._draw_type = draw_type

    @property
    def dtype(self)->GLInfo.attr_types:
        if self._dtype is not None:
            return self._dtype
        
        if self:
            return type(self._list[0])

        return None
    
    @dtype.setter
    def dtype(self, dtype:GLInfo.attr_types):
        str_dtype = str(dtype)
        if str_dtype == "callback_vec3":
            dtype = glm.vec3
        elif str_dtype == "callback_vec4":
            dtype = glm.vec4
        elif str_dtype == "callback_quat":
            dtype = glm.quat

        if self._dtype == dtype:
            return
        
        self._dtype = dtype
        self._list_dirty = True
        self._should_retest = True
        self.stride = sizeof(dtype)

    def _apply(self)->None:
        self._check_in_items()
        if self._increment is None:
            self.__first_apply()
        elif self._increment.is_changed:
            self.__apply_increment()

    def __first_apply(self)->None:
        if not self:
            return

        assert self._increment is None
        self._increment = Increment(self)

        assert self._vbo.nbytes == 0 and self._vbo.id == 0
        self.stride = sizeof(self.const_get(0))
        self._vbo.malloc(self.capacity*self.stride, self._draw_type)

        value_array = self.ndarray
        self._vbo.bufferSubData(0, value_array.nbytes, value_array)

        self.is_new_vbo = True

    def __apply_increment(self)->None:
        self.is_new_vbo = False

        assert self._increment.is_changed
        patch = self._increment.patch()
        if patch["old_capacity"] == patch["new_capacity"]:
            for move in patch["move"]:
                self._vbo.memmove(move["old_start"]*self.stride, move["size"]*self.stride, move["new_start"]*self.stride)
        else:
            temp_vbo = VBO()
            temp_vbo.malloc(patch["new_capacity"]*self.stride, self._draw_type)
            self._vbo.copy_to(0, min(patch["old_size"], patch["new_size"])*self.stride, temp_vbo, 0)
            for move in patch["move"]:
                self._vbo.copy_to(move["old_start"]*self.stride, move["size"]*self.stride, temp_vbo, move["new_start"]*self.stride)

            self._vbo.delete()
            self._vbo = temp_vbo
            self.is_new_vbo = True

        new_data = patch["new_data"]
        patch_update = patch["update"]

        if new_data and patch_update:
            temp_buffer = np.array(new_data)
            if len(patch_update) > 1:
                temp_vbo = VBO()
                temp_vbo.bufferData(temp_buffer, self._draw_type)
                for update in patch_update:
                    dest_start = update["dest_start"]
                    size = update["size"]
                    src_start = update["src_start"]
                    temp_vbo.copy_to(src_start*self.stride, size*self.stride, self._vbo, dest_start*self.stride)
            else:
                update = patch_update[0]
                dest_start = update["dest_start"]
                size = update["size"]
                self._vbo.bufferSubData(dest_start*self.stride, size*self.stride, temp_buffer)


import cv2
import os
from PIL import Image
import numpy as np
import platform
import importlib

from .utils import extname, printable_path, is_url, md5s
from .download import is_China_user, pip_install, download
from .GlassConfig import GlassConfig

OpenEXR = None
Imath = None

def import_OpenEXR()->None:
    global OpenEXR
    global Imath

    if OpenEXR is not None:
        return

    try:
        OpenEXR = importlib.import_module("OpenEXR")
        Imath = importlib.import_module("Imath")
    except:
        version = ".".join(platform.python_version().split(".")[:2])
        plat = "x64" if platform.architecture()[0] == "64bit" else "win32"
        if version == "3.12":
            github_url = ""
            gitee_url = ""

            if plat == "x64":
                github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/OpenEXR/OpenEXR-1.3.9-cp312-cp312-win_amd64.whl"
                gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/OpenEXR/OpenEXR-1.3.9-cp312-cp312-win_amd64.whl"
            else:
                github_url = "https://raw.githubusercontent.com/Time-Coder/Glass-Engine/main/OpenEXR/OpenEXR-1.3.9-cp312-cp312-win32.whl"
                gitee_url = "https://gitee.com/time-coder/Glass-Engine/raw/main/OpenEXR/OpenEXR-1.3.9-cp312-cp312-win32.whl"

            if is_China_user():
                pip_install(gitee_url)
            else:
                pip_install(github_url)
        else:
            openexr_urls = \
            {
                ("3.6", "win32"): "https://download.lfd.uci.edu/pythonlibs/archived/cp36/OpenEXR-1.3.2-cp36-cp36m-win32.whl",
                ("3.7", "win32"): "https://download.lfd.uci.edu/pythonlibs/archived/cp37/OpenEXR-1.3.7-cp37-cp37m-win32.whl",
                ("3.8", "win32"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp38-cp38-win32.whl",
                ("3.9", "win32"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp39-cp39-win32.whl",
                ("3.10", "win32"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp310-cp310-win32.whl",
                ("3.11", "win32"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp311-cp311-win32.whl",

                ("3.6", "x64"): "https://download.lfd.uci.edu/pythonlibs/archived/cp36/OpenEXR-1.3.2-cp36-cp36m-win_amd64.whl",
                ("3.7", "x64"): "https://download.lfd.uci.edu/pythonlibs/archived/cp37/OpenEXR-1.3.7-cp37-cp37m-win_amd64.whl",
                ("3.8", "x64"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp38-cp38-win_amd64.whl",
                ("3.9", "x64"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp39-cp39-win_amd64.whl",
                ("3.10", "x64"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp310-cp310-win_amd64.whl",
                ("3.11", "x64"): "https://download.lfd.uci.edu/pythonlibs/archived/OpenEXR-1.3.8-cp311-cp311-win_amd64.whl"
            }
            pip_install(openexr_urls[(version, plat)])
        
        OpenEXR = importlib.import_module("OpenEXR")
        Imath = importlib.import_module("Imath")

class ImageLoader:

    __image_map = {}

    @staticmethod
    def load(file_name:str):
        if not os.path.isfile(file_name):
            if is_url(file_name):
                url = file_name
                file_name = GlassConfig.cache_folder + "/" + md5s(url) + "/" + os.path.basename(url)
                if not os.path.isfile(file_name):
                    download(url, file_name)
            else:
                raise FileNotFoundError("not a valid image file: " + file_name)
        
        file_name = os.path.abspath(file_name).replace("\\", "/")
        if file_name in ImageLoader.__image_map:
            return ImageLoader.__image_map[file_name]
        ext_name = extname(file_name)
        if ext_name == "glsl":
            raise ValueError(f"not supported image format: '{file_name}'")
        
        if GlassConfig.print:
            print(f"loading image: {printable_path(file_name)} ", end="", flush=True)
        image = None
        if ext_name == "exr":
            image = ImageLoader.OpenEXR_load(file_name)
        else:
            image = ImageLoader.cv2_load(file_name)
            if image is None:
                image = ImageLoader.PIL_load(file_name)

        if image is None:
            if GlassConfig.print:
                print("")
            raise ValueError(f"not supported image format: '{file_name}'")
        
        image = cv2.flip(image, 0)

        if image.dtype in [np.float32, np.float64] and image.max() > 1:
            image = ImageLoader.tone_mapping(image)

        ImageLoader.__image_map[file_name] = image

        if GlassConfig.print:
            print("done")

        return image
    
    @staticmethod
    def cv2_load(file_name):
        try:
            image = cv2.imread(file_name, cv2.IMREAD_UNCHANGED)
            if image is None:
                return None
            
            channels = image.shape[2] if len(image.shape) > 2 else 1
            if channels == 3:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            elif channels == 4:
                image = cv2.cvtColor(image, cv2.COLOR_BGRA2RGBA)

            return image
        except:
            return None

    @staticmethod
    def PIL_load(file_name):
        try:
            pil_image = Image.open(file_name)
            should_convert = False
            dest_mode = pil_image.mode
            if pil_image.mode == "CMYK":
                dest_mode = "RGBA"
                should_convert = True
            elif pil_image.mode == "P":
                dest_mode = "RGBA"
                should_convert = True
            elif pil_image.mode == "1":
                dest_mode = "L"
                should_convert = True

            if should_convert:
                pil_image = pil_image.convert(dest_mode)

            return np.array(pil_image)
        except:
            return None
    
    @staticmethod
    def OpenEXR_load(file_name):
        import_OpenEXR()
        pt = Imath.PixelType(Imath.PixelType.FLOAT)
        img_exr = OpenEXR.InputFile(file_name)
        dw = img_exr.header()['dataWindow']
        shape = (dw.max.y - dw.min.y + 1, dw.max.x - dw.min.x + 1)
        r_bytes, g_bytes, b_bytes = img_exr.channels('RGB', pt)
        r = np.frombuffer(r_bytes, dtype=np.float32).reshape(shape)
        g = np.frombuffer(g_bytes, dtype=np.float32).reshape(shape)
        b = np.frombuffer(b_bytes, dtype=np.float32).reshape(shape)

        return cv2.merge((r, g, b))
    
    @staticmethod
    def tone_mapping(image):
        image[image < 0] = 0
        gray = np.copy(image)

        channels = image.shape[2] if len(image.shape) > 2 else 1
        if channels == 3:
            gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
        elif channels == 4:
            gray = cv2.cvtColor(image, cv2.COLOR_RGBA2GRAY)

        gray[gray>1] = 1 + np.tanh(gray[gray>1]-1)

        rows = gray.shape[0]
        cols = gray.shape[1]

        resize_rows = 64
        resize_cols = int(resize_rows / rows * cols)

        blur_rows = 3
        blur_radius = int(blur_rows / rows * cols)
        
        temp_gray = cv2.copyMakeBorder(gray, rows, rows, cols, cols, borderType=cv2.BORDER_WRAP)
        temp_gray = cv2.resize(temp_gray, (3*resize_cols, 3*resize_rows), interpolation=cv2.INTER_AREA)
        temp_gray = cv2.GaussianBlur(temp_gray, (2*blur_radius+1, 2*blur_radius+1), 0)
        temp_gray = cv2.resize(temp_gray, (3*cols, 3*rows))

        L = temp_gray[rows:-rows, cols:-cols]
        L = np.sqrt(L)

        if channels == 3 or channels == 4:
            L = cv2.merge((L, L, L))
            image[:,:,:3] = image[:,:,:3]/L / (1 + image[:,:,:3]/L)
            image[:,:,:3] /= image[:,:,:3].max()
            image = np.sin(np.pi/2*image[:,:,:3])
        else:
            image = image/L / (1 + image/L)
            image /= image.max()
            image = np.sin(np.pi/2*image)

        return image


from glass import GLConfig, ShaderProgram
import importlib

DirLight = None
PointLight = None
SpotLight = None
Material = None
Fog = None

def import_DirLight():
    global DirLight
    if DirLight is None:
        DirLight = importlib.import_module(".Lights.DirLight", "glass_engine").DirLight

def import_PointLight():
    global PointLight
    if PointLight is None:
        PointLight = importlib.import_module(".Lights.PointLight", "glass_engine").PointLight

def import_SpotLight():
    global SpotLight
    if SpotLight is None:
        SpotLight = importlib.import_module(".Lights.SpotLight", "glass_engine").SpotLight

def import_Material():
    global Material
    if Material is None:
        Material = importlib.import_module(".Material", "glass_engine").Material

def import_Fog():
    global Fog
    if Fog is None:
        Fog = importlib.import_module(".Fog", "glass_engine").Fog

class GlassEngineConfig:

    def __init__(self):
        self._has_material_recv_shadows:bool = False
        self._has_material_cast_shadows:bool = False
        self._has_material_recv_fog:bool = False

        self._has_dir_lights:bool = False
        self._has_point_lights:bool = False
        self._has_spot_lights:bool = False

        self._has_dir_lights_generate_shadows:bool = False
        self._has_point_lights_generate_shadows:bool = False
        self._has_spot_lights_generate_shadows:bool = False

        self._has_fog:bool = False

        self._dict = {}
        self._dict["USE_DYNAMIC_ENV_MAPPING"] = False
        self._dict["USE_DIR_LIGHT"] = False
        self._dict["USE_DIR_LIGHT_SHADOW"] = False
        self._dict["USE_POINT_LIGHT"] = False
        self._dict["USE_POINT_LIGHT_SHADOW"] = False
        self._dict["USE_SPOT_LIGHT"] = False
        self._dict["USE_SPOT_LIGHT_SHADOW"] = False
        self._dict["USE_FOG"] = False
        self._dict["USE_SHADING_MODEL_FLAT"] = False
        self._dict["USE_SHADING_MODEL_GOURAUD"] = False
        self._dict["USE_SHADING_MODEL_PHONG"] = False
        self._dict["USE_SHADING_MODEL_PHONG_BLINN"] = False
        self._dict["USE_SHADING_MODEL_TOON"] = False
        self._dict["USE_SHADING_MODEL_OREN_NAYAR"] = False
        self._dict["USE_SHADING_MODEL_MINNAERT"] = False
        self._dict["USE_SHADING_MODEL_COOK_TORRANCE"] = False
        self._dict["USE_SHADING_MODEL_UNLIT"] = False
        self._dict["USE_SHADING_MODEL_FRESNEL"] = False
        self._dict["USE_BINDLESS_TEXTURE"] = False

    def __getitem__(self, name:str):
        return self._dict[name]
    
    def __setitem__(self, name:str, value:bool):
        if name not in self._dict:
            return

        if self._dict[name] == value:
            return
        
        if not value:
            for program in ShaderProgram.all_instances:
                if self.__has_macro(program):
                    return
            self._dict[name] = value
        else:
            self._dict[name] = value
            for program in ShaderProgram.all_instances:
                self.__reload_program(program)
    
    def define_for_program(self, program):
        has_bindless_extension = ("GL_ARB_bindless_texture" in GLConfig.available_extensions)
        is_valid = False
        for name in self._dict:
            if name in ["USE_DYNAMIC_ENV_MAPPING", "USE_DIR_LIGHT_SHADOW", "USE_POINT_LIGHT_SHADOW", "USE_SPOT_LIGHT_SHADOW"] and \
               not has_bindless_extension:
                is_valid = program.undef(name) or is_valid
                continue

            if name == "USE_BINDLESS_TEXTURE":
                if has_bindless_extension:
                    is_valid = program.define(name, 1) or is_valid
                else:
                    is_valid = program.undef(name) or is_valid
                continue

            if self._dict[name]:
                is_valid = program.define(name, 1) or is_valid
            else:
                is_valid = program.undef(name) or is_valid
        return is_valid

    def __has_macro(self, program):
        for name in program.defines:
            if name in self._dict:
                return True

        return False

    def __reload_program(self, program):
        if not self.__has_macro(program):
            return
        
        if self.define_for_program(program):
            program.reload()

    def _update_dynamic_env(self, flag:bool):
        import_Material()

        has_dynamic_env_mapping = flag
        if not flag:
            for material in Material.all_instances:
                if material.dynamic_env_mapping:
                    has_dynamic_env_mapping = True
                    break
        
        self["USE_DYNAMIC_ENV_MAPPING"] = has_dynamic_env_mapping

    def _update_recv_shadows(self, flag:bool):
        import_Material()

        has_material_recv_shadows = flag
        if not flag:
            for material in Material.all_instances:
                if material.recv_shadows:
                    has_material_recv_shadows = True
                    break
        
        if self._has_material_recv_shadows == has_material_recv_shadows:
            return

        self._has_material_recv_shadows = has_material_recv_shadows
        if self._has_material_recv_shadows:
            if self._has_material_cast_shadows:
                if self._has_dir_lights_generate_shadows:
                    self["USE_DIR_LIGHT_SHADOW"] = True
                if self._has_point_lights_generate_shadows:
                    self["USE_POINT_LIGHT_SHADOW"] = True
                if self._has_spot_lights_generate_shadows:
                    self["USE_SPOT_LIGHT_SHADOW"] = True
        else:
            self["USE_DIR_LIGHT_SHADOW"] = False
            self["USE_POINT_LIGHT_SHADOW"] = False
            self["USE_SPOT_LIGHT_SHADOW"] = False

    def _update_cast_shadows(self, flag:bool):
        import_Material()

        has_material_cast_shadows = flag
        if not flag:
            for material in Material.all_instances:
                if material.cast_shadows:
                    has_material_cast_shadows = True
                    break
        
        if self._has_material_cast_shadows == has_material_cast_shadows:
            return

        self._has_material_cast_shadows = has_material_cast_shadows
        if self._has_material_cast_shadows:
            if self._has_material_recv_shadows:
                if self._has_dir_lights_generate_shadows:
                    self["USE_DIR_LIGHT_SHADOW"] = True
                if self._has_point_lights_generate_shadows:
                    self["USE_POINT_LIGHT_SHADOW"] = True
                if self._has_spot_lights_generate_shadows:
                    self["USE_SPOT_LIGHT_SHADOW"] = True
        else:
            self["USE_DIR_LIGHT_SHADOW"] = False
            self["USE_POINT_LIGHT_SHADOW"] = False
            self["UES_SPOT_LIGHT_SHADOW"] = False

    def _update_recv_fog(self, flag:bool):
        import_Material()

        has_material_recv_fog = flag
        if not flag:
            for material in Material.all_instances:
                if material.fog:
                    has_material_recv_fog = True
                    break
        
        if self._has_material_recv_fog == has_material_recv_fog:
            return

        self._has_material_recv_fog = has_material_recv_fog
        if self._has_material_recv_fog:
            if self._has_fog:
                self["USE_FOG"] = True
        else:
            self["USE_FOG"] = False

    def _update_fog(self):
        import_Fog()

        has_fog = False
        for fog in Fog.all_instances:
            if fog.extinction_density > 1E-6 and \
               fog.inscattering_density > 1E-6:
                has_fog = True
                break

        if self._has_fog == has_fog:
            return

        self._has_fog = has_fog
        if self._has_fog:
            if self._has_material_recv_fog:
                self["USE_FOG"] = True
        else:
            self["USE_FOG"] = False

    def _update_dir_lights(self):
        import_DirLight()

        self["USE_DIR_LIGHT"] = bool(DirLight.all_instances)

    def _update_point_lights(self):
        import_PointLight()

        self["USE_POINT_LIGHT"] = bool(PointLight.all_instances)

    def _update_spot_lights(self):
        import_SpotLight()

        self["USE_SPOT_LIGHT"] = bool(SpotLight.all_instances)

    def _update_dir_lights_generate_shadows(self, flag:bool):
        import_DirLight()

        has_dir_lights_generate_shadows = flag
        if not flag:
            for dir_light in DirLight.all_instances:
                if dir_light.generate_shadows:
                    has_dir_lights_generate_shadows = True
                    break

        if self._has_dir_lights_generate_shadows == has_dir_lights_generate_shadows:
            return

        self._has_dir_lights_generate_shadows = has_dir_lights_generate_shadows
        if self._has_dir_lights_generate_shadows:
            if self._has_material_recv_shadows and self._has_material_cast_shadows:
                self["USE_DIR_LIGHT_SHADOW"] = True
        else:
            self["USE_DIR_LIGHT_SHADOW"] = False

    def _update_point_lights_generate_shadows(self, flag:bool):
        import_PointLight()

        has_point_lights_generate_shadows = flag
        if not flag:
            for point_light in DirLight.all_instances:
                if point_light.generate_shadows:
                    has_point_lights_generate_shadows = True
                    break

        if self._has_point_lights_generate_shadows == has_point_lights_generate_shadows:
            return

        self._has_point_lights_generate_shadows = has_point_lights_generate_shadows
        if self._has_point_lights_generate_shadows:
            if self._has_material_recv_shadows and self._has_material_cast_shadows:
                self["USE_POINT_LIGHT_SHADOW"] = True
        else:
            self["USE_POINT_LIGHT_SHADOW"] = False

    def _update_spot_lights_generate_shadows(self, flag:bool):
        import_SpotLight()

        has_spot_lights_generate_shadows = flag
        if not flag:
            for spot_light in DirLight.all_instances:
                if spot_light.generate_shadows:
                    has_spot_lights_generate_shadows = True
                    break

        if self._has_spot_lights_generate_shadows == has_spot_lights_generate_shadows:
            return

        self._has_spot_lights_generate_shadows = has_spot_lights_generate_shadows
        if self._has_spot_lights_generate_shadows:
            if self._has_material_recv_shadows and self._has_material_cast_shadows:
                self["USE_SPOT_LIGHT_SHADOW"] = True
        else:
            self["USE_SPOT_LIGHT_SHADOW"] = False

    def _update_shading_model(self):
        import_Material()

        has_Flat = False
        has_Gouraud = False
        has_Phong = False
        has_PhongBlinn = False
        has_Toon = False
        has_OrenNayar = False
        has_Minnaert = False
        has_CookTorrance = False
        has_Unlit = False
        has_Fresnel = False
        for material in Material.all_instances:
            if material.shading_model == Material.ShadingModel.Flat:
                has_Flat = True

            elif material.shading_model == Material.ShadingModel.Gouraud:
                has_Gouraud = True

            elif material.shading_model == Material.ShadingModel.Phong:
                has_Phong = True

            elif material.shading_model == Material.ShadingModel.PhongBlinn:
                has_PhongBlinn = True

            elif material.shading_model == Material.ShadingModel.Toon:
                has_Toon = True

            elif material.shading_model == Material.ShadingModel.OrenNayar:
                has_OrenNayar = True

            elif material.shading_model == Material.ShadingModel.Minnaert:
                has_Minnaert = True

            elif material.shading_model in [Material.ShadingModel.CookTorrance, Material.ShadingModel.PBR]:
                has_CookTorrance = True

            elif material.shading_model in [Material.ShadingModel.NoShading, Material.ShadingModel.Unlit]:
                has_Unlit = True

            elif material.shading_model == Material.ShadingModel.Fresnel:
                has_Fresnel = True

        self["USE_SHADING_MODEL_FLAT"] = has_Flat
        self["USE_SHADING_MODEL_GOURAUD"] = has_Gouraud
        self["USE_SHADING_MODEL_PHONG"] = has_Phong
        self["USE_SHADING_MODEL_PHONG_BLINN"] = has_PhongBlinn
        self["USE_SHADING_MODEL_TOON"] = has_Toon
        self["USE_SHADING_MODEL_OREN_NAYAR"] = has_OrenNayar
        self["USE_SHADING_MODEL_MINNAERT"] = has_Minnaert
        self["USE_SHADING_MODEL_COOK_TORRANCE"] = has_CookTorrance
        self["USE_SHADING_MODEL_UNLIT"] = has_Unlit
        self["USE_SHADING_MODEL_FRESNEL"] = has_Fresnel

GlassEngineConfig = GlassEngineConfig()


def is_space(char):
    return char in [' ', '\t', '\n', '\r', '\v', '\f']

def is_identifier_part(char):
    return 'a' <= char <= 'z' or 'A' <= char <= 'Z' or '0' <= char <= '9' or char in ['$', '_']

def cut(s, lef, rig, ind):
    return s[lef[ind]:rig[ind] + 1]

def safe_index_of(s, ch, ind):
    while ind != len(s) and not is_space(s[ind]):
        if s[ind] == ch:
            return ind
        ind += 1
    return -1

def process_trigraph(str):
    trigraphs = {
        '=': '#',
        '/': '\\',
        "'": '^',
        '(': '[',
        ')': ']',
        '<': '{',
        '>': '}',
        '!': '|',
        '-': '~'
    }
    for ch, replacement in trigraphs.items():
        ret = ''
        lst = 0
        pattern = "??" + ch
        while lst != len(str):
            ind = str.find(pattern, lst)
            if ind == -1:
                break
            ret += str[lst:ind] + replacement
            lst = ind + len(pattern)
        ret += str[lst:]
        str = ret
    return str

def process_line_break(str):
    lines = str.split('\n')
    ret = ''
    for line in lines:
        j = len(line)
        while j > 0 and is_space(line[j - 1]):
            j -= 1
        if line[j - 1:j] == '\\':
            ret += line[:j - 1]
        else:
            ret += line + '\n'
    if ret and ret[-1] == '\n':
        ret = ret[:-1]
    return ret

def process_multiline_comment_and_replace_mark(str):
    ret = ''
    string_begin, esc = None, False
    in_string = False
    lst = 0
    i = 0
    while i < len(str):
        if in_string:
            if not esc and str[i] == string_begin:
                in_string = False
        elif str[i] in ["'", '"']:
            in_string = True
            string_begin = str[i]
        elif i + 1 < len(str):
            if str[i:i + 2] == '/*':
                ret += str[lst:i] + ' '
                i = str.find('*/', i + 2) + 2
                lst = i
                if i == -1:
                    break
            elif str[i:i + 2] == '%:':
                ret += str[lst:i] + '#'
                i += 2
                lst = i
        esc = (str[i] == '\\')
        i += 1
    ret += str[lst:]
    return ret

def minifyc(str):
    return _compress(_compress(str))

def _compress(str):
    ret = ''
    lst = 0
    pattern_s = "R\"("
    pattern_t = ")\""
    while lst < len(str):
        ind = str.find(pattern_s, lst)
        if ind == -1:
            break
        ret += compress_single(str[lst:ind])
        rig = str.find(pattern_t, ind + len(pattern_s))
        if rig == -1:
            ret += str[ind:]
            lst = len(str)
            break
        lst = rig + len(pattern_t)
        ret += str[ind:lst]
    ret += compress_single(str[lst:])
    return ret


def compress_single(str):
    str = process_trigraph(str)
    str = process_line_break(str)
    str = process_multiline_comment_and_replace_mark(str)

    arr = str.split('\n')
    ret = ""
    force_newline = True
    last = '\0'
    for line in arr:
        if not line:
            continue

        i = 0
        lef, rig = [], []

        while True:
            while i < len(line) and is_space(line[i]):
                i += 1
            if i == len(line):
                break
            lef.append(i)
            in_string = False
            string_begin, esc = None, False
            break_lex = False
            while i < len(line) and (in_string or not is_space(line[i])):
                if in_string:
                    if not esc and line[i] == string_begin:
                        in_string = False
                elif line[i] in ['"', '\'']:
                    in_string = True
                    string_begin = line[i]
                elif line[i] == '/' and i < len(line) - 1 and line[i + 1] == '/':
                    rig.append(i - 1)
                    lef.append(i)
                    rig.append(len(line) - 1)
                    break_lex = True
                    break
                i += 1
                if i == len(line):
                    break
                esc = (line[i] == '\\')
            if break_lex:
                break
            rig.append(i - 1)

        n = len(lef)
        if not n:
            continue

        origin_force_newline = force_newline
        origin_last = last

        if last != '\0':
            if force_newline:
                ret += '\n'
            elif check(last, line[lef[0]]):
                ret += ' '

        last = line[rig[n - 1]]
        j = 0

        if line[lef[0]] == '#':
            if not force_newline:
                ret += '\n'
            force_newline = True
            if not line[lef[0]+1]:
                j = 1
                lef[1] -= 1
                line = line[:lef[1]] + '#' + line[lef[1]+1:]
            ret += cut(line, lef, rig, j)
            if line[lef[j] + 1] == 'd' and safe_index_of(line, '(', lef[j + 1]) == -1:
                j += 1
                ret += ' ' + cut(line, lef, rig, j)
                if j + 1 < n:
                    j += 1
                    ret += (' ' + cut(line, lef, rig, j))
                
            j += 1
        else:
            force_newline = False

        while j < n:
            if lef[j] + 2 <= len(line) and line[lef[j]:lef[j] + 2] == '//':
                if j == 0:
                    force_newline = origin_force_newline
                break
            if j and check(line[rig[j - 1]], line[lef[j]]):
                ret += ' '
            ret += cut(line, lef, rig, j)
            j += 1

        last = line[rig[j - 1]] if j else origin_last

    return ret

def check(a, b):
    return (is_identifier_part(a) and is_identifier_part(b)) or \
           ((a in [' + ', '-']) and a == b) or \
           (a == '/' and b == '*')



from OpenGL import GL
import warnings

from .GPUProgram import GPUProgram, LinkError, LinkWarning
from .Shaders import ComputeShader
from .GLConfig import GLConfig
from .GlassConfig import GlassConfig

class ComputeProgram(GPUProgram):

	def __init__(self)->None:
		GPUProgram.__init__(self)
		self.compute_shader = ComputeShader()
		self._work_group_size:tuple = None
		self._should_join:bool = False

	@staticmethod
	def _check_work_group_size(work_group_size:tuple)->None:
		max_work_group_size = GLConfig.max_compute_work_group_size
		if work_group_size[0] > max_work_group_size[0]:
			raise ValueError("x-dimension work group size should not be greater than " + str(max_work_group_size[0]) + ", " + str(work_group_size[0]) + " were given")
		if work_group_size[1] > max_work_group_size[1]:
			raise ValueError("y-dimension work group size should not be greater than " + str(max_work_group_size[1]) + ", " + str(work_group_size[1]) + " were given")
		if work_group_size[2] > max_work_group_size[2]:
			raise ValueError("z-dimension work group size should not be greater than " + str(max_work_group_size[2]) + ", " + str(work_group_size[2]) + " were given")
		
		max_work_group_invocations = GLConfig.max_compute_work_group_invocations
		work_group_invocations = work_group_size[0] * work_group_size[1] * work_group_size[2]
		if work_group_invocations > max_work_group_invocations:
			raise ValueError("work group invocation should not be greater than " + str(max_work_group_invocations) + ", " + str(work_group_invocations) + " were given")

	@staticmethod
	def _check_work_group_count(work_group_count:tuple)->None:
		max_work_group_count = GLConfig.max_compute_work_group_count
		if work_group_count[0] > max_work_group_count[0]:
			raise ValueError("x-dimension work group size should not be greater than " + str(max_work_group_count[0]) + ", " + str(work_group_count[0]) + " were given")
		if work_group_count[1] > max_work_group_count[1]:
			raise ValueError("y-dimension work group size should not be greater than " + str(max_work_group_count[1]) + ", " + str(work_group_count[1]) + " were given")
		if work_group_count[2] > max_work_group_count[2]:
			raise ValueError("z-dimension work group size should not be greater than " + str(max_work_group_count[2]) + ", " + str(work_group_count[2]) + " were given")

	def compile(self, file_name:str)->None:
		self.compute_shader = ComputeShader.load(file_name)

		self._work_group_size = self.compute_shader.work_group_side
		self._uniforms_info = self.compute_shader.uniforms_info
		self._uniform_blocks_info = self.compute_shader.uniform_blocks_info
		self._shader_storage_blocks_info = self.compute_shader._shader_storage_blocks_info
		self._structs_info = self.compute_shader._structs_info

		ComputeProgram._check_work_group_size(self._work_group_size)

		self._is_linked = False

	def _apply(self)->None:
		if not self._linked_but_not_applied:
			return
		
		if self._id == 0:
			self._id = GL.glCreateProgram()
			if self._id == 0:
				raise MemoryError("failed to create ComputeProgram")
			
		GL.glAttachShader(self._id, self.compute_shader._id)
		GL.glLinkProgram(self._id)

		if GlassConfig.debug:
			message_bytes = GL.glGetProgramInfoLog(self._id)
			message = message_bytes
			if isinstance(message_bytes, bytes):
				message = str(message_bytes, encoding="utf-8")

			error_messages, warning_messages = self._format_error_warning(message)
			if warning_messages and GlassConfig.warning:
				warnings.warn("\n" + "\n".join(warning_messages), category=LinkWarning)

			if error_messages:
				self._is_linked = False
				raise LinkError("\n" + "\n".join(error_messages))
		
		self._apply_uniform_blocks()
		self._apply_shader_storage_blocks()

		self._linked_but_not_applied = True

	def _link(self)->None:
		if self._is_linked:
			return

		if not self.compute_shader.is_compiled:
			raise RuntimeError("should compile compute shader before link")

		self._resolve_uniforms()
		self._resolve_uniform_blocks()
		self._resolve_shader_storage_blocks()

		self._is_linked = True
		self._linked_but_not_applied = True
		if GL.glCreateProgram:
			self._apply()

	def start_computing(self, x_work_group_count:int, y_work_group_count:int=1, z_work_group_count:int=1)->None:
		ComputeProgram._check_work_group_count((x_work_group_count, y_work_group_count, z_work_group_count))
		self.use()
		GL.glDispatchCompute(x_work_group_count, y_work_group_count, z_work_group_count)
		self._should_join = True

	def join(self, barrier_type=GL.GL_ALL_BARRIER_BITS)->None:
		if self._should_join:
			GL.glMemoryBarrier(barrier_type)
			self._should_join = False

	def compute(self, x_work_group_count:int, y_work_group_count:int=1, z_work_group_count:int=1, barrier=GL.GL_ALL_BARRIER_BITS)->None:
		self.start_computing(x_work_group_count, y_work_group_count, z_work_group_count)
		self.join(barrier)

	@property
	def work_group_size(self)->tuple:
		return self._work_group_size

	@property
	def related_files(self)->list:
		return [self.compute_shader._file_name]


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel, QComboBox
from PyQt6.QtCore import pyqtSignal, Qt

class ComboboxInput(QWidget):

    value_changed = pyqtSignal(object)

    def __init__(self, prompt:str, values:dict, value=None, parent:QWidget=None):
        QWidget.__init__(self, parent=parent)

        hlayout = QHBoxLayout()
        hlayout.setContentsMargins(0,0,0,0)
        hlayout.setAlignment(Qt.AlignmentFlag.AlignLeft)

        label = QLabel(prompt, self)
        hlayout.addWidget(label)

        self.__combobox = QComboBox(self)
        active_index = -1
        index = 0
        for key, _value in values.items():
            self.__combobox.addItem(key, _value)
            if value is not None and _value == value:
                active_index = index

            index += 1

        if active_index != -1:
            self.__combobox.setCurrentIndex(active_index)

        hlayout.addWidget(self.__combobox)
        self.setLayout(hlayout)

        self.__combobox.currentIndexChanged.connect(self.__index_changed)

    @property
    def value(self):
        return self.__combobox.currentData()
    
    @value.setter
    def value(self, value):
        for i in range(self.__combobox.count()):
            if value == self.__combobox.itemData(i):
                self.__combobox.setCurrentIndex(i)
                return

    def __index_changed(self, index:int):
        self.value_changed.emit(self.__combobox.itemData(index))